#ifndef ALARM_H

#include <Arduino.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/io.h>
#include <avr/wdt.h>
#include <Time.h>
#define SEEED_ETHERNET 1
#if SEEED_ETHERNET
#include <EthernetV2_0.h>
#include <EthernetUdpV2_0.h>
#include <DnsV2_0.h>
#else
// #include <Ethernet.h>
// #include <EthernetUdp.h>
// #include <Dns.h>
#endif
#include <SPI.h>
#include "ntpstuff.h"

#include <HardwareSerial.h>
#include <EEPROM.h>
#include <Timezone.h>
// #include <DS1302RTC.h>

// I/O Connections:
#define EthResPin   2	// hardware Ethernet Reset
#define HWResetPin  2	// hardware everything Reset
#define AmpMute 0	// Audio amp mute (not used)
// 20,21 SDA,SCL -- battery clock, second LCD display
// 26,27,28,29,30,31,32,33 alarm contacts
#define FIRST_ALM_PIN 26
#define ALM_PIN_DELTA 1
#define ALARM_RELAY 34	// Alarm Relay
#define DNS_OK_LED  38
#define USE_DHCP_BUTTON 39
#define Wireless1   40	// wireless remote button 1
#define Wireless2   41	// wireless remote button 2
#define X10RTS      42
#define X10DTR      43
#define SFXBarking  44	// Adafruit Audio SFX Board Trigger 0
#define SFXNotes    45	// Adafruit Audio SFX Board Trigger 1
#define SFXTaDaaa   46	// Adafruit Audio SFX Board Trigger 2
#define SFXSadSound 47	// Adafruit Audio SFX Board Trigger 3
#define PRI_COL16SW     37
#define SEC_COL16SW     48
#define SECOND_UNIT_SW 49
#define SECOND_POT_A  7
#define SECOND_POT_B  6	// see also SECOND_E
#define SECOND_BKLT 12
#define SECOND_A    A13
#define SECOND_B    A12
#define SECOND_C    A11
#define SECOND_D    A10
#define SECOND_E    A2	// quad pot push button, 2nd front panel

#define MAXZONES 8
// front, back, sliding, basement outer,
// basement inner, master bedroom, PIR, garage

#define DS18S20DEBUG 0
#define EthDEBUG 1
#define EthHardCoded 0
#define NTPDEBUG 1
#define TimerDEBUG 0
#define GET_CLOCK_DEBUG 1
#define SDDEBUG 0
#define UIDEBUG 0
#define KBDDEBUG 0
#define BLDEBUG 0
#define EthRESETinSetup 0
#define EthRESET 0
#define WEB_IN_DEBUG 0	// show web server debug on console.
#define WEB_OUT_DEBUG 0	// show web server debug on console.

#define SDCARD 0

#define SDCARD_CS 4
// SS,MOSI,MISO,SCLK,CS
// 10,11,12,13,4 used by ethernet shield and sdcard.  (UNO)
// 53,51,50,52,4 used by ethernet shield and sdcard.  (Mega)
#define rxSense 11

#define ON 1
#define OFF 0

#if IS_MAIN
    #define EXTERN
    #define INIT(x) =x
    #define INIT3(x,y,z) =x,y,z
#else
    #define EXTERN extern
    #define INIT(x)
    #define INIT3(x,y,z)
#endif

#define MAXMSTIMERS 5
#define KBUITIMER 0
#define TOD_TIMER 1
#define ASCAN_TMR 2
#define X10_ACTION_TIMER 3
#define SOUND_TMR 4
EXTERN unsigned long msec_Timers[MAXMSTIMERS];

// first MAXZONES timers are for entry/exit delay, one per zone
#define MAXSTIMERS 6 + MAXZONES
#define MISC_TIMER MAXZONES + 0
#define SETCLOCK_TIMER MAXZONES + 1
#define GETCLOCK_TIMER MAXZONES + 2
#define EXIT_WCLOSED_TIMER MAXZONES + 3
#define LCD_UPDATE_TIMER MAXZONES + 4
#define SEND_EMAIL_TIMER MAXZONES + 5
EXTERN unsigned long sec_Timers[MAXSTIMERS];
#define MAXMTIMERS 5
#define X10_CHECK_TIMER 1
#define NTP_RETRY_TIMER 2
#define EXIT_WOPEN_TIMER 3
#define EEPROM_SAVE_TIMER 4
#define SCANDELAY 50
#define ASCAN_DEL 50
EXTERN unsigned long min_Timers[MAXMTIMERS];
EXTERN time_t tNow;
EXTERN char Dbuf[250];  //general purpose buffer for strings
EXTERN char Dbuf2[250];	// all the rest for building messages with sprintf
EXTERN char Dbuf4[20];
EXTERN unsigned char vweek, vdays, vhour, vmin;
EXTERN uint8_t ntp_fail_counter;
#define MAX_NTP_FAILURES 10

#define MAXDISPLAYS 2

typedef struct displ_t1
{
  unsigned int backlight_timer;
  unsigned char turn_off_backlight;
  unsigned char turn_on_backlight;
} CONFIG_DISP_RAM;

EXTERN CONFIG_DISP_RAM display[MAXDISPLAYS];

typedef struct displ_t2
{
  unsigned char rows;
  unsigned char cols;
} CONFIG_DISP_EEP;

#define CONFIG_REVISION_LEVEL 12335L

#define INPUT_SEX 0	// 1 for low to alarm, 0 for loop connection to ground
typedef struct config_t
{
  uint8_t enabled;	// 1 for enabled, 0 for disabled
  uint8_t contact;	// input pin for zone
  char name[10];
  unsigned int entry_delay_sec;	// delay in seconds
  unsigned int exit_delay_sec;	// delay in seconds	NOT USED
  uint8_t attributes;
} CONFIG_ZONE;

#define ATTR_NORMAL 1	// normal entryway
#define ATTR_OUTSIDE 2	// door to outside, monitor in NIGHT mode

typedef struct config_t2
{
    CONFIG_ZONE zone_config[MAXZONES];
    char password[4][11];
    char alarm_zone;
    uint8_t night_start[2];	// hh:mm in [1]:[0]
    uint8_t night_stop[2];	// hh:mm in [1]:[0]
    uint8_t night_mode_enabled;
} CONFIG_ALARM;
EXTERN CONFIG_ALARM alarm_config;

#define MAX_HOUSE_CODES 2
#define MAX_X10EVENTS 64

typedef struct action
{
    uint8_t evidx;
    uint8_t action;
    uint8_t ctr;
} ACTION;
EXTERN ACTION action_queue[MAX_X10EVENTS];

// note: total eeprom is 4096 bytes
// one event occupies 7 bytes
// room for a bit over 400 events
#define MAX_AEVENTS 400
typedef struct x10event
{
    uint8_t hour;		// 0..23, + 32,33 for sunrise, sunset
    uint8_t uidx;		// index into lamp_names table
    uint8_t action;		// off, on, no-op
} X10EVENT;

typedef struct aevent
{
    uint8_t sstate;	// system_state
    unsigned int openv;	// open_vector
    time_t ts;		// timestamp, GMT
} AEVENT;

typedef struct x10device
{
    char name[17];
    uint8_t is_inside_light;
} ADEVICE;

EXTERN unsigned int alm_evt_idx;
// this only exists in eeprom, not enough room in RAM:
typedef struct
{
    AEVENT alarm_events[MAX_AEVENTS];
} ALARM_LOG;
#define ADDR_ALARM_LOG (4095 - sizeof(ALARM_LOG))

enum TZindex { Eastern, Central, Mountain, Arizona, Pacific };

// IN EEPROM, order is:
// EEPROM_writeAnything(0, alarm_config);
// EEPROM_writeAnything(sizeof(CONFIG_ALARM), config_data);
// EEPROM_writeAnything(sizeof(CONFIG_DATA)+sizeof(CONFIG_ALARM), system_state);

typedef struct config_thing
{
    long magic_number;
    unsigned char house_codes[MAX_HOUSE_CODES];	// REAL DATA
    /* these are in order: F1, F2, F3, F4, . . . F16, H1, H2, . . . H16 */
    ADEVICE x10devices[MAX_HOUSE_CODES * 16];
    X10EVENT x10events[MAX_X10EVENTS];
    TZindex tzindex;
    uint8_t ip[4];
    uint8_t dns_server[4];
    uint8_t gateway_ip[4];
    uint8_t subnet_mask[4];
    uint8_t mac[6];
    char wifi_ssid[30];
    char wifi_pass[30];
    unsigned char backlight;
    unsigned char dimbacklight;
    unsigned char clock_is_set;	// if battery clock has been set
    unsigned int batt_cal_constant;	// battery calibration, ~1535
    CONFIG_DISP_EEP display[MAXDISPLAYS];
} CONFIG_DATA;
EXTERN CONFIG_DATA config_data;

EXTERN unsigned char datalog_done;
EXTERN unsigned char UIstate;
EXTERN unsigned char prog_zone;
EXTERN unsigned char prog_field;
// EXTERN unsigned char last_keypress;
// EXTERN unsigned char any_key_hit;
EXTERN unsigned char has_displayed;
EXTERN char Status[MAXZONES][8];
EXTERN unsigned char dns_working;
EXTERN unsigned char button_E_pressed;

#define ACT_FLD 0
#define HOUR_FLD 1
#define MIN_FLD 2
#define TEMP_FLD 3
#define MAX_FIELDS 4
EXTERN unsigned char blinking[MAXDISPLAYS][4];	// active, hour, min, temp

#define BLTIMEOUT 300	// backlight timeout (in seconds)
#define DIMBACKLIGHT 0x10

#define RUN   0
#define PROG1 1

#define DL(x) Serial.print(x)
#define DLn(x) Serial.println(x)
#define DLF(x) Serial.print(F(x))
#define DLFn(x) Serial.println(F(x))
#define DV(m, v) do{Serial.print(m);Serial.print(v);Serial.print(' ');}while(0)
#define DVn(m, v) do{Serial.print(m);Serial.println(v);}while(0)
#define DVF(m, v) do{Serial.print(F(m));Serial.print(' ');Serial.print(v);Serial.print(F(" "));}while(0)
#define DHF(m, v) do{Serial.print(F(m));Serial.print(' ');Serial.print(v, HEX);Serial.print(' ');}while(0)
#define DVFn(m, v) do{Serial.print(F(m));Serial.print(' ');Serial.println(v);}while(0)
#define DHFn(m, v) do{Serial.print(F(m));Serial.print(' ');Serial.println(v, HEX);}while(0)

time_t timelocal(time_t);
void resetFunc(void);

#define ALARM_H 1
#endif
