/* Web enabled alarm and X10 controller.  Of course it's open source.
 * William F. Dudley Jr.
 * 2017 09 26
 *
 * board: Arduino Mega 2560
 * processor Atmega 2560
 * programmer: AVRISP
   
    todo:
   
    nil
    
    pin usage:
   
    Analog  0
    Analog  1
    Analog  2	2nd control panel quadrature potentiometer push button
    Analog  . . .
    Analog  7
    Analog  8	battery voltage
    Analog 10	2nd control panel D button
    Analog 11	2nd control panel C button
    Analog 12	2nd control panel B button
    Analog 13	2nd control panel A button
    Analog 14	my WDT enable
    Analog 15	my WDT reset
   
    Digital 0 Serial input from USB interface
    Digital 1 Serial output to USB interface
    Digital 4 SD card select on ethernet board
    Digital 6 2nd control panel quadrature potentiometer B
    Digital 7 2nd control panel quadrature potentiometer A
PINOUT
     2 EthernetReset/HardwareReset
    16 TX2 to secondary LCD
    12 Backlight secondary LCD
    18,19 TX1,RX1 to primary LCD/UI
    20,21 SDA,SCL (II2) to local (secondary) display
    24,25 SDA,SCL to DS3231 clock
    26,27,28,29,30,31,32,33 alarm contacts
    34 alarm relay
OPEN 35, 36
    38 DNS ok LED
    39 use DHCP button
    40,41 wireless button 1 and 2
    42,43 x10 transmitter (RTS, DTR)
// RTS -> pin 7 on DB9
// DTR -> pin 4 on DB9
// GROUND pin 5 on DB9
    44,45,46,47 sound effects triggers
    48 20/16 column switch
    49 backup development unit switch
    50,51,52,53 Arduino Ethernet SPI interface
*/

// #define EthResPin   2	// hardware Ethernet Reset
// #define HWResetPin  2	// hardware everything Reset
// #define AmpMute 0	// Audio amp mute (not used)
// // 26,27,28,29,30,31,32,33 alarm contacts
// #define FIRST_ALM_PIN 26
// #define ALM_PIN_DELTA 1
// #define ALARM_RELAY 34	// Alarm Relay
// #define DNS_OK_LED  38
// #define USE_DHCP_BUTTON 39
// #define Wireless1   40	// wireless remote button 1
// #define Wireless2   41	// wireless remote button 2
// #define X10RTS      42
// #define X10DTR      43
// #define SFXBarking  44	// Adafruit Audio SFX Board Trigger 0
// #define SFXNotes    45	// Adafruit Audio SFX Board Trigger 1
// #define SFXTaDaaa   46	// Adafruit Audio SFX Board Trigger 2
// #define SFXSadSound 47	// Adafruit Audio SFX Board Trigger 3
// #define PRI_COL16SW     37
// #define SEC_COL16SW     48
// #define SECOND_UNIT_SW 49
// #define SECOND_POT_A  7
// #define SECOND_POT_B  6
// #define SECOND_BKLT 12
// #define SECOND_A    A13
// #define SECOND_B    A12
// #define SECOND_C    A11
// #define SECOND_D    A10
// #define SECOND_E    A2	// quad pot push button, 2nd front panel

#define SERDELAY delay(10); // serial display freezes or misses chars if < 10

#define IS_MAIN 1
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/io.h>
#if INTERNAL_WDT
// #include <avr/wdt.h>
#else
#include "wdt.h"
#endif
#include <Time.h>
#include <SPI.h>

#include <HardwareSerial.h>
#include <EEPROM.h>
#include <Timezone.h>

#include "EEPROM_rw_anything.h"
#include "alarm.h"
#if SEEED_ETHERNET
#include <EthernetV2_0.h>
#include <EthernetUdpV2_0.h>
#include <DnsV2_0.h>
#else
#include <Ethernet.h>
#include <EthernetUdp.h>
#include <Dns.h>
#endif
#include "WebServer.h"
#define PROFILING 0
#define PROFILING_MAIN 1
#include "profiling.h"
#include "lcdisplay.h"

#include "sunrise.h"
#include "ds3231.h"
#include <X10Firecracker.h>
#include "ui_mode.h"
#include "sound_types.h"
#include "sys_state.h"
#include <Wire.h>
#include "LiquidCrystal_I2C.h"
#include <ESP8266.h>
#include <Quadrature.h>

// this hack fixes the "truncation . . . bad_interrupt" problem.
// If the problem returns, dick with the size (0xf3e) a little.
// uint8_t PROGMEM a[0xf3e] = {1};

// #if INTERNAL_WDT :
// watchdog timer values:
// Supported on ATMega 8, 168, 328, 1280, 2560
// WDTO_15MS, WDTO_30MS, WDTO_60MS, WDTO_120MS,
//     WDTO_250MS, WDTO_500MS, WDTO_1S, WDTO_2S
// Supported on ATMega 168, 328, 1280, 2560
// WDTO_4S, WDTO_8S
// #endif

DS3231 rtc;


// Initialize the Ethernet client library
// with the IP address and port of the server
// that you want to connect to (port 80 is default for HTTP):

EthernetServer Control(80);

LiquidCrystal_I2C lcd(0x3f, 2, 1, 0, 4, 5, 6, 7, 3, POSITIVE);
#define BIGDISPLAY
#ifdef BIGDISPLAY
#define I2CCOLS 20
#define I2CROWS 4
#else
#define I2CCOLS 16
#define I2CROWS 2
#endif

#if SDCARD
const int SDchipSelect = 4;	// SD card chip select
void SDrecord (void);
#endif

TimeChangeRule usEDT = {"EDT", Second, Sun, Mar, 2, -240};  //UTC - 4 hours
TimeChangeRule usEST = {"EST", First, Sun, Nov, 2, -300};   //UTC - 5 hours
Timezone usEastern(usEDT, usEST);

//US Central Time Zone (Chicago, Houston)
TimeChangeRule usCDT = {"CDT", Second, dowSunday, Mar, 2, -300};
TimeChangeRule usCST = {"CST", First, dowSunday, Nov, 2, -360};
Timezone usCentral(usCDT, usCST);

//US Mountain Time Zone (Denver, Salt Lake City)
TimeChangeRule usMDT = {"MDT", Second, dowSunday, Mar, 2, -360};
TimeChangeRule usMST = {"MST", First, dowSunday, Nov, 2, -420};
Timezone usMountain(usMDT, usMST);

//Arizona is US Mountain Time Zone but does not use DST
Timezone usArizona(usMST, usMST);

//US Pacific Time Zone (Las Vegas, Los Angeles)
TimeChangeRule usPDT = {"PDT", Second, dowSunday, Mar, 2, -420};
TimeChangeRule usPST = {"PST", First, dowSunday, Nov, 2, -480};
Timezone usPacific(usPDT, usPST);

#define InterruptOff  TIMSK2 &= ~(1<<TOIE2)
#define InterruptOn  TIMSK2 |= (1<<TOIE2)
volatile unsigned int int_counter;
volatile unsigned char seconds;
volatile unsigned char new_minute;
volatile unsigned long uptime_min;
unsigned int tcnt2; // used to store timer value
volatile unsigned char perform_x10_action;

// front, back, sliding, basement outer, basement inner, (master) bedroom, PIR, garage
const char front_door[] PROGMEM = "Front";
const char kitchen_door[] PROGMEM = "Kitchen";
const char sliding_door[] PROGMEM = "Sliding";
const char basement_outer_door[] PROGMEM = "Back Door";
const char basement_inner_door[] PROGMEM = "BaseStair";
const char master_door[] PROGMEM = "Bedroom";
const char pir[] PROGMEM = "PIR";
const char garage[] PROGMEM = "Garage";
#define LONGEST_NAME 10	// longest one including the trailing null

const char * const zone_names[] PROGMEM = {
    front_door,
    kitchen_door,
    sliding_door,
    basement_outer_door,
    basement_inner_door,
    master_door,
    pir,
    garage
};

// ATTR_NORMAL = 0, 1, 6
// ATTR_OUTSIDE = 0, 1, 2, 3, 7

const char sh_front_door[] PROGMEM = "Frt";
const char sh_kitchen_door[] PROGMEM = "Kit";
const char sh_sliding_door[] PROGMEM = "Sld";
const char sh_basement_outer_door[] PROGMEM = "Bck";
const char sh_basement_inner_door[] PROGMEM = "Bas";
const char sh_master_door[] PROGMEM = "Bed";
const char sh_pir[] PROGMEM = "PIR";
const char sh_garage[] PROGMEM = "Gar";

const char * const short_zone_names[] PROGMEM = {
    sh_front_door,
    sh_kitchen_door,
    sh_sliding_door,
    sh_basement_outer_door,
    sh_basement_inner_door,
    sh_master_door,
    pir,
    sh_garage
};

char passwd[2][11], last_num[2];
uint8_t current_zone[2];	// used by UI_ZSET code; 0=ILLEGAL
uint8_t pwp[2];	// password pointer
uint8_t open[MAXZONES];	// "door is open" indicator
uint8_t last_open[MAXZONES];	// "door was open" indicator
unsigned int open_vector;	// bitmapped "open" indicator, bit 1 = zone 0
unsigned int open_enabled_vector; // bitmapped "open" indicator, for enabled zones
uint8_t open_count;
uint8_t lcd_needs_update;
uint8_t email_failed_counter;
#define MAX_EMAIL_FAILS 3

uint8_t alert_zone;	// will be zone + 1 of zone causing alarm
uint8_t email_alert_zone;
uint8_t is_entry_delay;
uint8_t need_to_send_email;
int uil_idx[2], uil_cnt[2];	// one for each front panel
uint8_t uil_dir[2];
// for second front panel
uint16_t last_quad_pos_result;
uint8_t second_fp_char;
uint8_t first_poll_second_fp;

Quadrature quad1(SECOND_POT_B, SECOND_POT_A);

uint8_t backup_unit;

#define MAX_WIFI_TRIES 3

ESP8266 wifi(Serial3, 115200);
byte wifi_connected;
byte ntp_host_idx_wifi;
time_t getNTPtimeAny (void);
time_t getNTPtimeWiFi (void);
time_t get_NTP_timeWiFi (const char *);
uint32_t sendNTPpacketWiFi (const char *);
uint16_t average_batt_reading, average_batt_filter;

uint8_t arm_system (void);
void disarm_system (void);
void set_system_wait (void);
void alert_owner (uint8_t);	// arg is (zone + 1) where zone is causing alarm
void printlnIP (byte *);	// print IP address on debug console

// change server to your email server ip or domain
IPAddress mailserver (192, 168, 0, 10);
String mailserverWiFi = "192.168.0.10";
char mailclientname[] = "mailclient.somedomain.com";
char recipient_name[] = "email_target@a_domain.com";
char from_email_addr[] = "You <youremail@someisp.com>";
char alarm_location[] = "your street address";
EthernetClient client;
byte sendEmail (uint8_t);
byte sendEmailWiFi (uint8_t);
uint8_t wireless_state[2];
void release_the_hounds (void);
void corral_the_hounds (void);
void alarm_bell_on (void);
void alarm_bell_off (void);
uint8_t quick_scan_loop (void);
uint8_t try_set_backup_rtc (void);
void HandleDisplay(unsigned char);
void eeprom_save_all(void);
void eeprom_save_alarm_config(void);
void eeprom_save_config_data(void);
void eeprom_restore_state(void);
void eeprom_save_state(void);
void eeprom_save_alm_idx(void);
void eeprom_restore_alm_idx(void);
void log_alarm(void);
void show_alarm_log(int);
void set_system_wclosed (void);
void send_command(char, unsigned char, unsigned char);
byte eRcv(void);

// no-cost stream operator as described at 
// http://sundial.org/arduino/?page_id=119
template<class T> inline Print &operator <<(Print &obj, T arg)
{ obj.print(arg); return obj; }

#define DEBOUNCE 2	// number of scan_loop() calls for a zone to change

// this is in sound_types.h:
// typedef enum { MUTE, BARKING, NOTES, TADAAA, SADSOUND } Sound_Type ;
const byte idx2port[] PROGMEM
    { AmpMute, SFXBarking, SFXNotes, SFXTaDaaa, SFXSadSound };
const unsigned int idx2time[] { 0L, 0L, 700L, 420L, 1500L };
byte sound_needs_mute;

void play_sound (Sound_Type sound_idx) {
    if(sound_idx == MUTE) {
	digitalWrite(SFXBarking, HIGH);	// no sound
	digitalWrite(SFXNotes, HIGH);	// no sound
	digitalWrite(SFXTaDaaa, HIGH);	// no sound
	digitalWrite(SFXSadSound, HIGH);	// no sound
#if AmpMute
	digitalWrite(AmpMute, LOW);	// mute amp
#endif
    }
    else {
#if AmpMute
	digitalWrite(AmpMute, HIGH);	// un-mute
#endif
	digitalWrite(pgm_read_byte(&idx2port[sound_idx]), LOW);
#if USE_PROGMEM_FOR_THIS
    unsigned long sndelay;
	memcpy_P((void *)&sndelay, (char*)pgm_read_word(&(idx2time[sound_idx])), sizeof(unsigned long));
	DVFn("sndelay =", sndelay);
	if(sndelay) {
	    msec_Timers[SOUND_TMR] = sndelay;
	    sound_needs_mute = 1;
	}
#else
	if(idx2time[sound_idx]) {
	    msec_Timers[SOUND_TMR] = (unsigned long)idx2time[sound_idx];
	    sound_needs_mute = 1;
	}
#endif
    }
}

void release_the_hounds (void) {
    DLFn("SFXBarking LOW");
    play_sound(BARKING);
}

void corral_the_hounds (void) {
    DLFn("SFXBarking HIGH");
    play_sound(MUTE);
}

void set_ui_mode (UI_Mode_Type uim, uint8_t dis) {
    last_num[dis] = '\0';
    current_zone[dis] = 0;
    memset((void *)passwd[dis], '\0', sizeof(passwd)/2);
    pwp[dis] = 0;
    DVFn("set_ui_mode()", (uint8_t)uim);
    if(uim == ui_mode[dis]) { return; }
    ui_mode[dis] = uim;
    if(uim == UI_IDLE
    || uim == UI_ARMED
    || uim == UI_LOG) {
	if(dis == 0) {
	    LCDquadpotMin (0, 0);
	    LCDquadpotMax (0, 9);
	}
	else {
	    quad1.minimum(0);
	    quad1.maximum(9);
	}
    }
    else if(uim == UI_ZSET) {
	if(dis == backup_unit) {
	    LCDquadpotMin (0, 1);
	    LCDquadpotMax (0, MAXZONES);
	}
	else {
	    quad1.minimum(1);
	    quad1.maximum(MAXZONES);
	}
    }
    HandleDisplay(dis);
}

// Aruino runs at 16 Mhz, so we have 1000 Overflows per second...
// this ISR will get hit once a millisecond
ISR(TIMER2_OVF_vect) {
unsigned char zone;

    quad1.poll();
    int_counter++;
    if (int_counter == 1000) {
	seconds++;
	int_counter = 0;
	for(uint8_t zone = 0 ; zone < MAXSTIMERS ; zone++) {
	    if(sec_Timers[zone]) {
		sec_Timers[zone]--;
		if(!sec_Timers[zone] && zone < MAXZONES) {
		    if(is_entry_delay) {
			alert_zone = zone + 1;
			is_entry_delay = 0;
		    }
		}
		if(!sec_Timers[zone] && zone == LCD_UPDATE_TIMER) {
		    lcd_needs_update++;
		    sec_Timers[LCD_UPDATE_TIMER] = 59;
		}
	    }
	}
	for(uint8_t i = 0 ; i < MAXDISPLAYS ; i++) {
	    if(display[i].backlight_timer) {
		display[i].backlight_timer--;
		if(!display[i].backlight_timer) {
		    if(!i) { DLFn("set turn_off_backlight"); }
		    display[i].turn_off_backlight = 1;
		}
	    }
	}
	if(seconds == 60) {
	    new_minute = 1;
	    seconds = 0;
	    uptime_min++;
	}
    }
    for(byte i = 0 ; i < MAXMSTIMERS ; i++) {
	if(msec_Timers[i]) { msec_Timers[i]--; }
    }
    if(!msec_Timers[X10_ACTION_TIMER]) {
	msec_Timers[X10_ACTION_TIMER] = 1060;
	perform_x10_action = 1;
    }
#if PROFILING
    prof_array[prof_line]++;
#endif
    TCNT2 = tcnt2;  // reset the timer for next time
}

// Timer setup code borrowed from Sebastian Wallin
// http://popdevelop.com/2010/04/mastering-timer-interrupts-on-the-arduino/
// This was the best explanation I found on the web.

void setupTimer() {
  //Timer2 Settings:  Timer Prescaler /1024
  // First disable the timer overflow interrupt while we're configuring
  TIMSK2 &= ~(1<<TOIE2);
  // Configure timer2 in normal mode (pure counting, no PWM etc.)
  TCCR2A &= ~((1<<WGM21) | (1<<WGM20));
  // Select clock source: internal I/O clock
  ASSR &= ~(1<<AS2);
  // Disable Compare Match A interrupt enable (only want overflow)
  TIMSK2 &= ~(1<<OCIE2A);

  // Now configure the prescaler to CPU clock divided by 128
  TCCR2B |= (1<<CS22)  | (1<<CS20); // Set bits
  TCCR2B &= ~(1<<CS21);             // Clear bit

  /* We need to calculate a proper value to load the timer counter.
   * The following loads the value 131 into the Timer 2 counter register
   * The math behind this is:
   * (CPU frequency) / (prescaler value) = 125000 Hz = 8us.
   * (desired period) / 8us = 125.
   * MAX(uint8) - 125 = 131;
   */
  /* Save value globally for later reload in ISR */
  tcnt2 = 131;

  /* Finally load end enable the timer */
  TCNT2 = tcnt2;
  TIMSK2 |= (1<<TOIE2);
  sei();
}

void lcd_show_open_zones(uint8_t di, uint8_t lcdrow, uint8_t cols) {
HardwareSerial *hsp;
char zname[21];
    if(open_enabled_vector) {
	hsp = LCDportarray[di];
	memset(zname, '\0', sizeof(zname));
	for(uint8_t zone = 0 ; zone < MAXZONES ; zone++) {
	    if(open[zone]) {
		char thiszname[LONGEST_NAME];
		strncpy_P(thiszname, (char*)pgm_read_word(&((open_count < 5) ? zone_names[zone] : short_zone_names[zone])), LONGEST_NAME);
		if((strlen(zname) + strlen(thiszname)) > cols) {
		    LCDline(di, lcdrow++);
		    (*hsp).print(zname); SERDELAY;
		    if(di == 1) lcd.print(zname);   
		    memset(zname, '\0', sizeof(zname));
		}
		// DVFn("zname =", zname);
		strncat(zname, thiszname, sizeof(zname)-strlen(zname));
		if(strlen(zname) < cols) {
		    strcat(zname, " ");
		}
	    }
	}
	if(strlen(zname)) {	// if unprinted list of short zone names
	    LCDline(di, lcdrow);
	    (*hsp).print(zname); SERDELAY;
	    if(di == 1) lcd.print(zname);   
	}
    }
}

// format string showing battery voltage and WiFi status
char * FormatBatteryVoltage (char *sname) {
    unsigned int v = ((long)average_batt_reading * (long)config_data.batt_cal_constant) >> 10;
    sprintf_P(sname, PSTR("Batt: %d.%02dV WiFi "), v/100, v%100);
    strcat(sname, (wifi_connected) ? "OK" : "ng");
}

// format string showing uptime
char * PrintUpTime_Dbuf2 (uint8_t cols) {
    // uptime_min = 1440 + 120 + 3;  // 1 day, 2:03 for testing
    // must compute these here, sprintf_P cannot handle the
    // calculations done in the argument list.
    unsigned int updays = uptime_min/1440L ;
    unsigned int uphrs = (uptime_min % 1440L)/60 ;
    unsigned int upmins = uptime_min % 60L ;
    // DVF("uptime_min =", uptime_min);
    // this breaks if I try to display four integers:
    sprintf_P(Dbuf2, PSTR("=> %d days %02d:%02d"), updays, uphrs, upmins);
    Serial.println(Dbuf2);
    sprintf_P(Dbuf2, (cols == 20) ? PSTR("%4d days %02d:%02d IDLE") : PSTR("%4dd %02d:%02d IDLE"), updays, uphrs, upmins);
    return(Dbuf2);
}

// zone is which LCD to use, datazone is which zone's data to display
void HandleDisplay(unsigned char di) {
HardwareSerial *hsp;
unsigned char j;
uint8_t cols;
char sname[22];

    lcd_needs_update = 0;
    PF(3);
    if(di == 0 && last_system_state != system_state) {
	strncpy_P(sname, (char*)pgm_read_word(&(sstate_names[(int)system_state])), sizeof(sname));
	DVFn("HD: system_state=", sname);
	DVFn("HD: ui_mode[di]=", ui_mode[di]);
	last_system_state = system_state;
    }
    hsp = LCDportarray[di];
    if(di == 0) {
	cols = (!digitalRead(PRI_COL16SW)) ? 16 : config_data.display[di].cols;
    }
    else {
	cols = (!digitalRead(SEC_COL16SW)) ? 16 : config_data.display[di].cols;
    }
    if(ui_mode[di] == UI_IDLE) {
    /*--1234567890      IDLE
	zone1----|zone2-----	// show zone if door is open, blank otherwise
	zone3----|zone4-----
	zone5----|zone6-----
     */
	LCDclr(di);
	if(digitalRead(USE_DHCP_BUTTON) && !button_E_pressed) {  // not pressed
	    sprintf_P(Dbuf2, (cols == 20) ? PSTR("  %-10.10s    IDLE") : PSTR("  %-9.9s IDLE"), passwd[di]);
	}
	else {
	    PrintUpTime_Dbuf2(cols);
	    DLn(Dbuf2);
	}
	(*hsp).print(Dbuf2); SERDELAY;
	if(di == 1) lcd.print(Dbuf2);
	if(digitalRead(USE_DHCP_BUTTON) && !button_E_pressed) {  // not pressed
	    lcd_show_open_zones(di, 1, cols);
	}
	else {
	    FormatBatteryVoltage(sname);
	    LCDline(di, 1);
	    DLn(sname);
	    (*hsp).print(sname);
	    if(di == 1) lcd.print(sname);
	    button_E_pressed--;
	}
	// put date and time on last line
	LCDline(di, 3);
	sprintf_P(Dbuf,PSTR("%02d/%02d/%4d %02d:%02d"),month(tNow),day(tNow),year(tNow),hour(tNow),minute(tNow));
	(*hsp).print(Dbuf); SERDELAY;
	if(di == 1) lcd.print(Dbuf);
    }
    else if(ui_mode[di] == UI_ARMED) {
    /*--1234567890     ARMED
     */
	LCDclr(di);
	sprintf_P(Dbuf2, (cols == 20) ? PSTR("%-10.10s     ARMED") : PSTR("%-10.10s ARMED"), passwd[di]);
	(*hsp).print(Dbuf2); SERDELAY;
	if(di == 1) lcd.print(Dbuf2);
	LCDline(di, 1);
	if(system_state == WOPEN) {
	    if(cols == 20) {
		(*hsp).print(F("Waiting: door open "));
		if(di == 1) lcd.print(F("Waiting: door open "));
	    }
	    else {
		(*hsp).print(F("Waiting: open  "));
		if(di == 1) lcd.print(F("Waiting: open  "));
	    }
	    SERDELAY;
	}
	else if(system_state == WCLOSED) {
	    if(cols == 20) {
		(*hsp).print(F("Waiting: door close"));
		if(di == 1) lcd.print(F("Waiting: door close"));
	    }
	    else {
		(*hsp).print(F("Waiting: closed"));
		if(di == 1) lcd.print(F("Waiting: closed"));
	    }
	    SERDELAY;
	}
	else if(system_state == ARMED) {
	    // show open zones on lines 1 (and 2, if needed)
	    // IS THIS A GOOD THING OR NOT ?
	    // lcd_show_open_zones(di, 1, cols);
	    // show date and time on line 3
	    LCDline(di, 3);
	    sprintf_P(Dbuf,PSTR("%02d/%02d/%4d %02d:%02d"),month(tNow),day(tNow),year(tNow),hour(tNow),minute(tNow));
	    (*hsp).print(Dbuf); SERDELAY;
	    if(di == 1) lcd.print(Dbuf);
	}
	else if(system_state == ALERT) {
	    if(cols == 20) {
		(*hsp).print(F(" ALARM: CALLING 911 ")); SERDELAY;
		if(di == 1) lcd.print(F(" ALARM: CALLING 911 "));
		LCDline(di, 2);
		(*hsp).print(F("   CALLING OWNER    ")); SERDELAY;
		if(di == 1) lcd.print(F("   CALLING OWNER    "));
		LCDline(di, 3);
		(*hsp).print(F("RELEASING THE HOUNDS"));
		if(di == 1) lcd.print(F("RELEASING THE HOUNDS"));
	    }
	    else {
		(*hsp).print(F("  CALLING 911  ")); SERDELAY;
		if(di == 1) lcd.print(F("  CALLING 911  "));
		LCDline(di, 2);
		(*hsp).print(F(" CALLING OWNER ")); SERDELAY;
		if(di == 1) lcd.print(F(" CALLING OWNER "));
		LCDline(di, 3);
		(*hsp).print(F("RELEASING DOGS "));
		if(di == 1) lcd.print(F("RELEASING DOGS "));
	    }
	    SERDELAY;
	}
    }
    else if(ui_mode[di] == UI_ZSET) {
	LCDclr(di);
	sprintf_P(Dbuf2, (cols == 20) ? PSTR("%-10.10s ZONE CTRL") : PSTR("%-9.9s ZONSET"), (current_zone[di] == 0) ? "TURN KNOB" : alarm_config.zone_config[current_zone[di]-1].name);
	(*hsp).print(Dbuf2); SERDELAY;
	if(di == 1) lcd.print(Dbuf2);
	if(current_zone[di]) {
	    LCDline(di, 1);
	    sprintf_P(Dbuf2, PSTR("%-10.10s          "), alarm_config.zone_config[current_zone[di]-1].enabled ? "Enabled" : "Disabled");
	    (*hsp).print(Dbuf2); SERDELAY;
	    if(di == 1) lcd.print(Dbuf2);
	}
    }
    else if(ui_mode[di] == UI_LOG) {
    time_t gmtNow, lts;
    AEVENT a;
    unsigned int eeaddr;
    char sname[10];
    char zname[21];
    uint8_t goodlog;
	LCDclr(di);
	gmtNow = now();
	do {
	    // DVF("uil_cnt =", uil_cnt[di]); DVFn("uil_idx =", uil_idx[di]);
	    goodlog = 1;
	    eeaddr = ADDR_ALARM_LOG + (uil_idx[di] * sizeof(AEVENT));
	    EEPROM_readAnything(eeaddr, a);
	    if(a.ts > gmtNow || a.ts < 1455000000L) {
		if(uil_dir[di]) {	// forwards in time
		    uil_idx[di]++;
		    uil_idx[di] %= MAX_AEVENTS;
		    uil_cnt[di]--;
		    if(uil_cnt[di] < 0) { uil_cnt[di] += MAX_AEVENTS; }
		}
		else {	// "normal" backwards in time
		    uil_idx[di]--;
		    if(uil_idx[di] < 0) { uil_idx[di] += MAX_AEVENTS; }
		    uil_cnt[di]++;
		    uil_cnt[di] %= MAX_AEVENTS;
		}
		goodlog = 0;
	    }
	} while(!goodlog);
	lts = timelocal(a.ts);
	strncpy_P(sname, (char*)pgm_read_word(&(sstate_names[(int)a.sstate])), sizeof(sname));
	sprintf_P(Dbuf, (cols == 20) ? PSTR("%8s %3d     LOG") : PSTR("%8s %3d LOG"), sname, uil_cnt[di]);
	if(cols == 20) {
	    sprintf_P(Dbuf2,PSTR("%02d/%02d/%4d %02d:%02d:%02d"),month(lts),day(lts),year(lts),hour(lts),minute(lts),second(lts));
	}
	else {
	    sprintf_P(Dbuf2,PSTR("%02d/%02d/%02d %02d:%02d:%01d"),month(lts),day(lts),year(lts) % 100,hour(lts),minute(lts),second(lts)/10);
	}
	(*hsp).print(Dbuf); SERDELAY;
	if(di == 1) lcd.print(Dbuf);
	LCDline(di, 1);
	(*hsp).print(Dbuf2); SERDELAY;
	if(di == 1) lcd.print(Dbuf2);
	if(a.openv) {
	uint8_t lcdrow = 2;
	uint8_t fillcnt = 0;
	    memset(zname, '\0', sizeof(zname));
	    for(uint8_t zone = 0 ; zone < MAXZONES ; zone++) {
		// DVFn("zone =", zone);
		if(a.openv & 1) {
		    // DVFn("active zone =", zone);
		    strncat_P(zname, (char*)pgm_read_word(&(short_zone_names[zone])), sizeof(zname) - strlen(zname));
		    strcat(zname, " ");
		    // DVFn("zname =", zname);
		    fillcnt++;
		    if(fillcnt == 4) {
			LCDline(di, lcdrow++);
			(*hsp).print(zname); SERDELAY;
			if(di == 1) lcd.print(zname);   
			fillcnt = 0;
			memset(zname, '\0', sizeof(zname));
		    }
		}
		a.openv >>= 1;
	    }
	    if(fillcnt) {	// if unprinted list of short zone names
		LCDline(di, lcdrow);
		(*hsp).print(zname); SERDELAY;
		if(di == 1) lcd.print(zname);   
	    }
	}
    }
    else if(ui_mode[di] == UI_NIGHT) {
    /*--1234567890     NIGHT
     */
	LCDclr(di);
	sprintf_P(Dbuf2, (cols == 20) ? PSTR("%-10.10s     NIGHT") : PSTR("%-10.10s NIGHT"), passwd[di]);
	(*hsp).print(Dbuf2); SERDELAY;
	if(di == 1) lcd.print(Dbuf2);
	LCDline(di, 1);
	if(system_state == NIGHT) {
	    // show date and time on line 3
	    LCDline(di, 3);
	    sprintf_P(Dbuf,PSTR("%02d/%02d/%4d %02d:%02d"),month(tNow),day(tNow),year(tNow),hour(tNow),minute(tNow));
	    (*hsp).print(Dbuf); SERDELAY;
	    if(di == 1) lcd.print(Dbuf);
	}
	else if(system_state == NITEALRT) {
	    if(cols == 20) {
		(*hsp).print(F("RELEASING THE HOUNDS"));
		if(di == 1) lcd.print(F("RELEASING THE HOUNDS"));
	    }
	    else {
		(*hsp).print(F("RELEASING DOGS "));
		if(di == 1) lcd.print(F("RELEASING DOGS "));
	    }
	    SERDELAY;
	}
    }
}

#ifdef HWResetPin	// hardware everything Reset
void resetFunc (void) {
    system_state = REBOOT;
    eeprom_save_all();
    digitalWrite(HWResetPin, HIGH);  // bang!
}
#else
void(* resetFunc) (void) = 0; //declare reset function @ address 0
#endif

void printlnIP (byte * ipa) {
    for(byte i = 0 ; i < 4 ; i++) {
	Serial.print(ipa[i]);
	Serial.print((i==3) ? '\n' : '.');
    }
}

void L1printslow (const char *s) {
char tmp[21];
    strncpy_P(tmp, s, 20);
    for(byte i = 0 ; tmp[i] ; i++) {
	LCDserial1.print(tmp[i]);
	delay(1);
    }
}

void eeprom_save_all(void) {
    eeprom_save_alarm_config();
    eeprom_save_config_data();
    eeprom_save_state();
    eeprom_save_alm_idx();
}

void eeprom_save_alm_idx (void) {
unsigned int idx;
  EEPROM_readAnything(ADDR_ALARM_LOG-2, idx);
  if(idx != alm_evt_idx) {
      EEPROM_writeAnything(ADDR_ALARM_LOG-2, alm_evt_idx);
  }
  // DLFn("alarm index saved.");
}

void eeprom_restore_alm_idx (void) {
  EEPROM_readAnything(ADDR_ALARM_LOG-2, alm_evt_idx);
}

void log_alarm (void) {
AEVENT a;
unsigned int eeaddr;
    alm_evt_idx++;
    alm_evt_idx %= MAX_AEVENTS;
    a.sstate = system_state;
    a.openv = open_enabled_vector;
    a.ts = now();
    eeaddr = ADDR_ALARM_LOG + (alm_evt_idx * sizeof(AEVENT));
    EEPROM_writeAnything(eeaddr, a);
    // for debug only:
    DLFn("-log_alarm()---- ALARM LOG BEGIN -----");
    show_alarm_log(10);
    DLFn("-log_alarm()----  ALARM LOG END  -----");
}

void show_alarm_log (int n) {
int idx;
time_t gmtNow, lts;
AEVENT a;
unsigned int eeaddr;
char sname[10];
char zname[33];
    gmtNow = now();
    idx = alm_evt_idx;
    for(unsigned int j = 0 ; j < n ; j++, --idx) {
	if(idx < 0) { idx = MAX_AEVENTS - 1; }
	eeaddr = ADDR_ALARM_LOG + (idx * sizeof(AEVENT));
	EEPROM_readAnything(eeaddr, a);
	if(a.ts > gmtNow) { continue; }	// in the future, can't be in log
	if(a.ts < 1455000000L) { continue; }	// before log feature starts
	lts = timelocal(a.ts);
	sprintf_P(Dbuf,PSTR("%02d/%02d/%4d %02d:%02d:%02d, "),month(lts),day(lts),year(lts),hour(lts),minute(lts),second(lts));
	Serial.print(Dbuf);
	strncpy_P(sname, (char*)pgm_read_word(&(sstate_names[(int)a.sstate])), sizeof(sname));
	Serial.print(sname);
	DLF(", ");
	if(a.openv) {
	    memset(zname, '\0', sizeof(zname));
	    for(uint8_t zone = 0 ; zone < MAXZONES ; zone++) {
		// DVFn("zone =", zone);
		if(a.openv & 1) {
		    // DVFn("active zone =", zone);
		    strncat_P(zname, (char*)pgm_read_word(&(short_zone_names[zone])), sizeof(zname) - strlen(zname));
		    strcat_P(zname, PSTR(" "));
		    // DVFn("zname =", zname);
		}
		a.openv >>= 1;
	    }
	    Serial.println(zname);
	}
	else {
	    DLFn("none open");
	}
    }
}

void eeprom_save_alarm_config(void) {
CONFIG_ALARM tempconfig;
    PF(6);
    EEPROM_readAnything(0, tempconfig); // get saved settings
    if(memcmp((const void *)&tempconfig, (const void *)&alarm_config, sizeof(CONFIG_ALARM))) {
	EEPROM_writeAnything(0, alarm_config); // save settings
    }
}

void eeprom_save_config_data(void) {
CONFIG_DATA tempconfig;
    PF(7);
    EEPROM_readAnything(sizeof(CONFIG_ALARM), tempconfig);
    if(memcmp((const void *)&tempconfig, (const void *)&config_data, sizeof(CONFIG_DATA))) {
	EEPROM_writeAnything(sizeof(CONFIG_ALARM), config_data);
    }
}

void eeprom_restore_state(void) {
    EEPROM_readAnything(sizeof(CONFIG_DATA)+sizeof(CONFIG_ALARM), system_state);
}

void eeprom_save_state(void) {
Sys_State_Type temp_state;
    EEPROM_readAnything(sizeof(CONFIG_DATA) + sizeof(CONFIG_ALARM), temp_state);
    if(temp_state != system_state) {
	EEPROM_writeAnything(sizeof(CONFIG_DATA)+sizeof(CONFIG_ALARM), system_state);
    }
}

#define wait_for_eth_HW 0

#if EthRESET
void ethernetReset (void) {
#if wait_for_eth_HW
  int cnt = 10, result;
#endif

  PF(7);
#if wait_for_eth_HW
  pinMode(rxSense, INPUT);  //for stabilizing the ethernet board
  while(1){
#endif
    digitalWrite(EthResPin, LOW);
    pinMode(EthResPin, OUTPUT);
    digitalWrite(EthResPin, HIGH);  // ethernet board reset
    mydelay(10);
    digitalWrite(EthResPin, LOW);
    mydelay(1000);
#if wait_for_eth_HW
  // now, after the reset, check the rx pin for constant on
    cnt = 10;
    result = 0;
    while (cnt--){  // count the number of times the light is on
      result += digitalRead(rxSense);
      mydelay(50);
    }
    Serial.println(result,DEC);
    if (result >=6)      // experimentation gave me this number YMMV
      return;
    mydelay(50);
  }
#endif
}
#endif

extern EthernetUDP Udp;
uint8_t action_ctr;

#if XXX_SEEED_ETHERNET
// EthernetServer webserver(80);
#else
#define PREFIX ""
// WebServer webserver(PREFIX, 80);
#endif
// void webserver_init(void);
void defaultCmd(WebServer &, WebServer::ConnectionType, char *, bool);
void houseCmd(WebServer &, WebServer::ConnectionType, char *, bool);
void unitCmd(WebServer &, WebServer::ConnectionType, char *, bool);
void schedCmd(WebServer &, WebServer::ConnectionType, char *, bool);
void manualCmd(WebServer &, WebServer::ConnectionType, char *, bool);
void configCmd(WebServer &, WebServer::ConnectionType, char *, bool);
void alarm_Cmd(WebServer &, WebServer::ConnectionType, char *, bool);
void almlog_Cmd(WebServer &, WebServer::ConnectionType, char *, bool);

void setup (void) {
unsigned char zone;
unsigned char sp;
// CONFIG_DATA test_config_data;
unsigned long test_revision;
byte config_needs_init;
int i;
    Serial.begin(115200);
    DVFn("sizeof(CONFIG_DATA) =", sizeof(CONFIG_DATA));
    DVFn("sizeof(CONFIG_ALARM) =", sizeof(CONFIG_ALARM));
    DVFn("sizeof(ALARM_LOG) =", sizeof(ALARM_LOG));
    DVFn("total eeprom =", sizeof(CONFIG_DATA)+sizeof(CONFIG_ALARM)+sizeof(ALARM_LOG)+3);
#if INTERNAL_WDT
    wdt_disable();
#else
    WDT_disable();
#endif
    Serial.println(F("Alarm System Initialization"));
#if PROFILING
    PF(0);
    prof_has_dumped = 0;
    clear_profiling_data();
#endif
#ifdef HWResetPin	// hardware everything Reset
    digitalWrite(HWResetPin, LOW);
    pinMode(HWResetPin, OUTPUT);
#endif
#if EthRESET
    pinMode(EthResPin, OUTPUT);
    digitalWrite(EthResPin, LOW);
#endif
    pinMode(SFXBarking, OUTPUT);	// Adafruit Audio SFX Board Trigger 0
    pinMode(SFXNotes, OUTPUT);		// Adafruit Audio SFX Board Trigger 1
    pinMode(SFXTaDaaa, OUTPUT);		// Adafruit Audio SFX Board Trigger 2
    pinMode(SFXSadSound, OUTPUT);	// Adafruit Audio SFX Board Trigger 3
#if AmpMute
    pinMode(AmpMute, OUTPUT);		// Amplifier Mute
#endif
    pinMode(PRI_COL16SW, INPUT);		// 20 col high, 16 col low
    digitalWrite(PRI_COL16SW, HIGH); // pullup
    pinMode(SEC_COL16SW, INPUT);		// 20 col high, 16 col low
    digitalWrite(SEC_COL16SW, HIGH); // pullup
    play_sound(MUTE);
    sound_needs_mute = 0;
    pinMode(ALARM_RELAY, OUTPUT);	// Alarm Relay
    alarm_bell_off();
    pinMode(Wireless1, INPUT);		// wireless remote button I
    pinMode(Wireless2, INPUT);		// wireless remote button II
    digitalWrite(Wireless1, HIGH); // pullup
    digitalWrite(Wireless2, HIGH); // pullup
    pinMode(DNS_OK_LED, OUTPUT);
    pinMode(USE_DHCP_BUTTON, INPUT);
    digitalWrite(USE_DHCP_BUTTON, HIGH);	// enable pullup
    pinMode(SECOND_UNIT_SW, INPUT);	// high if this is 2nd devel unit
    digitalWrite(SECOND_UNIT_SW, HIGH);	// enable pullup
    pinMode(SECOND_A, INPUT);
    digitalWrite(SECOND_A, HIGH); // enable pullup
    pinMode(SECOND_B, INPUT);
    digitalWrite(SECOND_B, HIGH); // enable pullup
    pinMode(SECOND_C, INPUT);
    digitalWrite(SECOND_C, HIGH); // enable pullup
    pinMode(SECOND_D, INPUT);
    digitalWrite(SECOND_D, HIGH); // enable pullup
    pinMode(SECOND_E, INPUT);
    digitalWrite(SECOND_E, HIGH); // enable pullup
    wireless_state[0] = 0;
    wireless_state[1] = 0;
    tNow = 0;
    int_counter = 0;
    seconds = 0;
    alert_zone = 0;
    email_alert_zone = 0;
    is_entry_delay = 0;
    need_to_send_email = 0;
    lcd_needs_update = 0;
    new_minute = 0;
    email_failed_counter = 0;
    ntp_fail_counter = 0;
    memset((void *)&open[0], 0, sizeof(open));
    memset((void *)&last_open[0], 0, sizeof(last_open));
    open_vector = 0;
    open_enabled_vector = 0;
    open_count = 0;
    last_system_state = ILLEGAL;
    sec_Timers[MISC_TIMER] = 60;

    // x10 config data is after alarm system config data
    config_needs_init = 0;
    // get saved non-alarm settings
    test_revision = 0L;
    DVFn("before, test_revision =", test_revision);
    i = EEPROM_readAnything(sizeof(CONFIG_ALARM), test_revision);
    DVFn("bytes read from EE =", i);
    DVFn("test_revision =", test_revision);
    DVFn("CONFIG_REVISION_LEVEL =", CONFIG_REVISION_LEVEL);
    EEPROM_readAnything(0, alarm_config); // get saved settings
    EEPROM_readAnything(sizeof(CONFIG_ALARM), config_data);
    DLF("alarm pins: ");
    for(zone = 0; zone < MAXZONES ; zone++) {
	Serial.print((int)alarm_config.zone_config[zone].contact);
	if(zone < (MAXZONES - 1)) { DLF(", "); }
	else { Serial.println(); }
    }
    if(test_revision != CONFIG_REVISION_LEVEL
    || config_data.ip[0] != 192) {
	DLn("config WRONG, writing config_data");
	config_needs_init = 1;
	memset((void *)&config_data.magic_number, '\0', sizeof(CONFIG_DATA));
	config_data.magic_number = CONFIG_REVISION_LEVEL;
	config_data.ip[0] = 192;	// 192.168.27.244 (dummy) or 248 (real)
	config_data.ip[1] = 168;
	config_data.ip[2] =  27;
	config_data.ip[3] = 248;
	config_data.gateway_ip[0] = config_data.ip[0];
	config_data.gateway_ip[1] = config_data.ip[1];
	config_data.gateway_ip[2] = config_data.ip[2];
	config_data.gateway_ip[3] = 1;
	config_data.dns_server[0] = 8;
	config_data.dns_server[1] = 8;
	config_data.dns_server[2] = 4;
	config_data.dns_server[3] = 4;
	config_data.subnet_mask[0] = 255;
	config_data.subnet_mask[1] = 255;
	config_data.subnet_mask[2] = 255;
	config_data.subnet_mask[3] =   0;
	config_data.mac[0] = 0xde;
	config_data.mac[1] = 0xad;
	config_data.mac[2] = 0xbe;
	config_data.mac[3] = 0xef;
	config_data.mac[4] = 0xfe;
	config_data.mac[5] = 0xc4;
	strcpy_P(config_data.wifi_ssid, PSTR("YOUR_SSID"));
	strcpy_P(config_data.wifi_pass, PSTR("WIFI_PWD"));
	config_data.tzindex = Eastern;
	config_data.house_codes[0] = 5;	// F
	config_data.house_codes[1] = 7;	// H
	memset(&config_data.x10devices[0], MAX_HOUSE_CODES * 16 * sizeof(ADEVICE), '\0');
	strcpy_P(config_data.x10devices[ 0].name, PSTR("south living rm"));
	strcpy_P(config_data.x10devices[ 1].name, PSTR("couch light"));
	strcpy_P(config_data.x10devices[ 2].name, PSTR("east living rm"));
	strcpy_P(config_data.x10devices[ 3].name, PSTR("north flood"));
	strcpy_P(config_data.x10devices[ 4].name, PSTR("southeast flood"));
	strcpy_P(config_data.x10devices[ 5].name, PSTR("west flood"));
	strcpy_P(config_data.x10devices[ 6].name, PSTR("garage flood"));
	strcpy_P(config_data.x10devices[ 7].name, PSTR("bathroom hw circ"));
	strcpy_P(config_data.x10devices[16].name, PSTR("sewing table lmp"));
	config_data.x10devices[ 0].is_inside_light = 1;
	config_data.x10devices[ 1].is_inside_light = 1;
	config_data.x10devices[ 2].is_inside_light = 1;
	config_data.x10devices[16].is_inside_light = 1;
	for(i = 32 ; i < MAX_X10EVENTS ; i++) {
	    config_data.x10events[i].action = 2;
	}
	i = 0;
	config_data.x10events[i].hour = 32;
	config_data.x10events[i].uidx = 4;
	// config_data.x10events[i].action = 0;
	config_data.x10events[++i].hour = 32;
	config_data.x10events[i].uidx = 5;
	// config_data.x10events[i].action = 0;
	config_data.x10events[++i].hour = 32;
	config_data.x10events[i].uidx = 6;
	// config_data.x10events[i].action = 0;
	config_data.x10events[++i].hour = 9;
	config_data.x10events[i].uidx = 4;
	// config_data.x10events[i].action = 0;
	config_data.x10events[++i].hour = 9;
	config_data.x10events[i].uidx = 5;
	// config_data.x10events[i].action = 0;
	config_data.x10events[++i].hour = 9;
	config_data.x10events[i].uidx = 6;
	// config_data.x10events[i].action = 0;
	config_data.x10events[++i].hour = 10;
	config_data.x10events[i].uidx = 4;
	// config_data.x10events[i].action = 0;
	config_data.x10events[++i].hour = 10;
	config_data.x10events[i].uidx = 5;
	// config_data.x10events[i].action = 0;
	config_data.x10events[++i].hour = 10;
	config_data.x10events[i].uidx = 6;
	// config_data.x10events[i].action = 0;
	config_data.x10events[++i].hour = 33;
	config_data.x10events[i].uidx = 0;
	config_data.x10events[i].action = 1;
	config_data.x10events[++i].hour = 33;
	config_data.x10events[i].uidx = 1;
	config_data.x10events[i].action = 1;
	config_data.x10events[++i].hour = 33;
	config_data.x10events[i].uidx = 4;
	config_data.x10events[i].action = 1;
	config_data.x10events[++i].hour = 33;
	config_data.x10events[i].uidx = 5;
	config_data.x10events[i].action = 1;
	config_data.x10events[++i].hour = 33;
	config_data.x10events[i].uidx = 6;
	config_data.x10events[i].action = 1;
	config_data.x10events[++i].hour = 0;
	config_data.x10events[i].uidx = 0;
	// config_data.x10events[i].action = 0;
	config_data.x10events[++i].hour = 0;
	config_data.x10events[i].uidx = 1;
	// config_data.x10events[i].action = 0;
	config_data.x10events[++i].hour = 0;
	config_data.x10events[i].uidx = 2;
	// config_data.x10events[i].action = 0;
	config_data.x10events[++i].hour = 1;
	config_data.x10events[i].uidx = 2;
	// config_data.x10events[i].action = 0;
	config_data.x10events[++i].hour = 0;
	config_data.x10events[i].uidx = 16;
	// config_data.x10events[i].action = 0;
	config_data.x10events[++i].hour = 1;
	config_data.x10events[i].uidx = 16;
	// config_data.x10events[i].action = 0;
	config_data.x10events[i].hour = 32;
	config_data.x10events[i].uidx = 3;
	config_data.x10events[i].action = 2;
	config_data.x10events[++i].hour = 9;
	config_data.x10events[i].uidx = 3;
	config_data.x10events[i].action = 2;
	config_data.x10events[++i].hour = 10;
	config_data.x10events[i].uidx = 3;
	config_data.x10events[i].action = 2;
	config_data.x10events[++i].hour = 22;
	config_data.x10events[i].uidx = 3;
	config_data.x10events[i].action = 2;
	config_data.x10events[++i].hour = 23;
	config_data.x10events[i].uidx = 3;
	config_data.x10events[i].action = 2;
	config_data.x10events[++i].hour = 0;
	config_data.x10events[i].uidx = 3;
	config_data.x10events[i].action = 2;
	config_data.x10events[++i].hour = 7;
	config_data.x10events[i].uidx = 7;
	config_data.x10events[i].action = 2;
	config_data.x10events[++i].hour = 10;
	config_data.x10events[i].uidx = 7;
	config_data.x10events[i].action = 2;
	config_data.x10events[++i].hour = 11;
	config_data.x10events[i].uidx = 7;
	config_data.x10events[i].action = 2;
	config_data.x10events[++i].hour = 20;
	config_data.x10events[i].uidx = 7;
	config_data.x10events[i].action = 2;
	config_data.x10events[++i].hour = 0;
	config_data.x10events[i].uidx = 7;
	config_data.x10events[i].action = 2;
	config_data.x10events[++i].hour = 1;
	config_data.x10events[i].uidx = 7;
	config_data.x10events[i].action = 2;
	config_data.backlight = 0xA0;
	config_data.dimbacklight = DIMBACKLIGHT;
	config_data.clock_is_set = 0;
	config_data.display[0].rows = 4;
	config_data.display[1].rows = 4;
	config_data.display[0].cols = digitalRead(PRI_COL16SW) ? 20 : 16;
	config_data.display[1].cols = digitalRead(SEC_COL16SW) ? 20 : 16;
	DVFn("cols=", config_data.display[0].cols);
	EEPROM_writeAnything(sizeof(CONFIG_ALARM), config_data);
    }
    for(zone = 0; zone < MAXZONES ; zone++) {
	sec_Timers[zone] = 60;
	/*
	blinking[zone][ACT_FLD] = '\0';
	blinking[zone][HOUR_FLD] = '\0';
	blinking[zone][MIN_FLD] = '\0';
	blinking[zone][TEMP_FLD] = '\0';
	*/
	if(config_needs_init) {
	    // DLn("inside init alarm_config");
	    alarm_config.zone_config[zone].enabled = 1;
	    // 26, 27, 28, 29, 30, 31, 32, 33
	    // alarm_config.zone_config[zone].contact = 26 + zone;
	    alarm_config.zone_config[zone].contact
		= FIRST_ALM_PIN + (zone << (ALM_PIN_DELTA-1));
	    // front or back get long delay
	    if(zone == 0 || zone == 1) {
		alarm_config.zone_config[zone].entry_delay_sec = 120; // 120 sec
		alarm_config.zone_config[zone].attributes = ATTR_NORMAL | ATTR_OUTSIDE;
	    }
	    else if(zone == 6) {	// PIR
		alarm_config.zone_config[zone].attributes = ATTR_NORMAL;
		alarm_config.zone_config[zone].entry_delay_sec = 120; // 120 sec
	    }
	    else {
		alarm_config.zone_config[zone].entry_delay_sec = 3; // 3 sec
		// set ATTR_OUTSIDE = zones 0, 1, 2, 3, 7
		if(zone == 2 || zone == 3 || zone == 7) {
		    alarm_config.zone_config[zone].attributes = ATTR_OUTSIDE;
		}
		else {
		    alarm_config.zone_config[zone].attributes = 0;
		}
	    }
	    strncpy_P(alarm_config.zone_config[zone].name, (char*)pgm_read_word(&(zone_names[zone])), sizeof(alarm_config.zone_config[zone].name));
	}
	pinMode(alarm_config.zone_config[zone].contact, INPUT);
	digitalWrite(alarm_config.zone_config[zone].contact, HIGH);	// pullup
    }
    // patch the config data to fix PIR response to normal house entry
    alarm_config.zone_config[6].attributes = ATTR_NORMAL;
    alarm_config.zone_config[6].entry_delay_sec = 120; // 120 sec
    if(config_needs_init) {
	alarm_config.alarm_zone = 0;
	DLn("inside init alarm_config.password");
	strcpy_P(alarm_config.password[0], PSTR("1234"));
	strcpy_P(alarm_config.password[1], PSTR("4321"));
	strcpy_P(alarm_config.password[2], PSTR(" "));
	strcpy_P(alarm_config.password[3], PSTR(" "));
	alarm_config.night_mode_enabled = 1;
	alarm_config.night_start[1] = 0;
	alarm_config.night_start[0] = 0;	// NIGHT mode at 00:00
	alarm_config.night_stop[1] = 7;
	alarm_config.night_stop[0] = 0;	// NIGHT mode ends at 07:00
	EEPROM_writeAnything(0, alarm_config); // save settings
    }
    if(config_data.batt_cal_constant < 1480
    || config_data.batt_cal_constant > 1600) {
	config_data.batt_cal_constant = (digitalRead(SECOND_UNIT_SW)) ? 1545 : 1533 ;
    }
    average_batt_reading = analogRead(A8);
    average_batt_filter = average_batt_reading << 4;
    Serial.println(F("Initializing LCD Displays"));
    setupLCD();
    lcd.begin(I2CCOLS, I2CROWS);

    for(i = 0 ; i < MAXDISPLAYS ; i++) {	// init backlights to "on"
	display[i].turn_off_backlight = 0;
	display[i].turn_on_backlight = 1;
	LCDbacklight(i, (unsigned int)config_data.backlight);
    }
    LCDclr(0);
    LCDserial1.print(F(" =-=-=-ALARM-=-=-= ")); SERDELAY;
    LCDclr(1);
    LCDserial2.print(F(" =-=-ALARM-=-= ")); SERDELAY;
    lcd.print(F(" =-=-=-ALARM-=-=-= "));
    LCDline(0, 1);
    LCDserial1.print(F("Booting...")); SERDELAY;
    LCDline(1, 1);
    LCDserial2.print(F("Booting...")); SERDELAY;
    lcd.print(F("Booting..."));
    // start the Ethernet connection:
    // first, reset the board and make sure it intialized correctly
#if EthRESETinSetup
#if EthDEBUG
    Serial.println(F("anti ethernetReset()"));
    LCDline(0, 1);
    // LCDserial1.print(F("anti ethReset()")); SERDELAY;
    L1printslow(PSTR("anti ethReset()"));
    LCDline(1, 1);
    LCDserial2.print(F("anti ethReset()")); SERDELAY;
    lcd.print(F("anti ethReset()"));
    delay(100);
#endif
    ethernetReset();
#if EthDEBUG
    DLFn("post ethernetReset()");
#endif
#endif
#if EthDEBUG
    DLFn("anti Ethernet.begin()");
    LCDclr(0);
    LCDline(0, 1);
    // LCDserial1.print(F("anti Eth.begin()")); SERDELAY;
    L1printslow(PSTR("anti eth.begin()"));
    LCDline(1, 1);
    LCDserial2.print(F("anti Eth.begin()"));
#endif
    pinMode(SDCARD_CS,OUTPUT);
    digitalWrite(SDCARD_CS,HIGH);//Deselect the SD card
    if(digitalRead(USE_DHCP_BUTTON)) {	// button not pressed
	uint8_t i;
	if(!digitalRead(SECOND_UNIT_SW) && config_data.ip[3] >= 245) {	// button pressed
	    config_data.mac[5]--;
	    config_data.ip[3] = 244;
	}
	DLF("Using static IP: ");
	printlnIP(config_data.ip);
	DLF("Using dns: ");
	printlnIP(config_data.dns_server);
	DLF("Using gateway: ");
	printlnIP(config_data.gateway_ip);
	DLF("Using subnet: ");
	printlnIP(config_data.subnet_mask);
	LCDline(0, 2);
	// LCDserial1.print(F("Using static IP: ")); SERDELAY;
	L1printslow(PSTR("Using static IP: "));
	LCDline(0, 3);
	LCDline(1, 2);
	LCDserial2.print(F("Using static IP: "));
	lcd.print(F("Using static IP: "));
	LCDline(1, 3);
	LCDprintIP(0,config_data.ip);
	LCDprintIP(1,config_data.ip);
	Ethernet.begin(config_data.mac, config_data.ip, config_data.dns_server, config_data.gateway_ip, config_data.subnet_mask);
	mydelay(1000);
    }
    else {
	DLFn("Using DHCP");
	if(Ethernet.begin(config_data.mac) == 0) {
	    DLFn("Failed to configure Ethernet using DHCP");
	    mydelay(10000);
	}
	else { // print your local IP address:
	    DLF("IP address:");
#if SEEED_ETHERNET
	    Serial.println(Ethernet.localIP());
#else
	    printlnIP(&Ethernet.localIP()[0]);
#endif
	    LCDline(0, 2);
	    LCDserial1.print(F("IP address:     ")); SERDELAY;
	    LCDline(2, 2);
	    LCDserial2.print(F("IP address:     "));
	    LCDline(0, 3);
	    LCDprintIP(0,&Ethernet.localIP()[0]);
	    LCDline(1, 3);
	    LCDprintIP(1,&Ethernet.localIP()[0]);
	}
	while(!digitalRead(USE_DHCP_BUTTON)) {	// while button is pressed
	    mydelay(100);  // wait for human to read message
	}
    }
#if EthDEBUG
    DLFn("post Ethernet.begin()");
    LCDline(0, 1);
    LCDserial1.print(F("post Eth.begin() ")); SERDELAY;
    LCDline(1, 1);
    LCDserial2.print(F("post Eth.begin() "));
    lcd.print(F("post Eth.begin() "));
#endif
    mydelay(1000);  // wait a bit for the card to stabilize
    if(1) {
	DNSClient dns;
        IPAddress remote_addr;
	int ret;
        DLFn("before dns.begin()");
	dns.begin(Ethernet.dnsServerIP());
        DLFn("before dns.getHostByName('pool.ntp.org')");
	ret = dns.getHostByName("pool.ntp.org", remote_addr);
	// ret = 0;
	if (ret == 1) {
	    DLF("address of pool.ntp.org: ");
	    printlnIP(&remote_addr[0]);
	    dns_working = 1;
	} else {
	    DLFn("(1) dns failed");
	    LCDline(0, 1);
	    LCDserial1.print(F("dns failed     ")); SERDELAY;
	    LCDline(1, 1);
	    LCDserial2.print(F("dns failed     "));
	    lcd.print(F("dns failed     "));
	    dns_working = 0;
	}
	mydelay(3000);
    }
    DVF("config.ssid =", config_data.wifi_ssid);
    DVFn(", config.pass =", config_data.wifi_pass);
    wifi_setup();
#if NTPDEBUG
    DLFn("time request");
#endif
    ntp_host_idx_wifi = ((byte)millis()) % (sizeof(ntp_hosts)/sizeof(const char *));
    setSyncInterval(1431 * 60);  //update time just over once a day
    setSyncProvider(getNTPtimeAny);

    Control.begin(); // start the web control server

#if TimerDEBUG
    Serial.println(F("before setupTimer()"));
#endif
    setupTimer();
#if TimerDEBUG
    Serial.println(F("post setupTimer()"));
#endif

    // make sure that the default chip select pin is set to
    // output, even if you don't use it:
    pinMode(SDCARD_CS, OUTPUT);
#if SDCARD
// SD Card stuff:
// On the Ethernet Shield, CS is pin 4. Note that even if it's not
// used as the CS pin, the hardware SS pin (10 on most Arduino boards,
// 53 on the Mega) must be left as an output or the SD library
// functions will not work.
#if SDDEBUG
    DLF("Initializing SD card...");
#endif

    // see if the card is present and can be initialized:
    if (!SD.begin(SDchipSelect)) {
#if SDDEBUG
	DLFn("Card failed, or not present");
#endif
    }
    else {
	DLFn("card initialized.");
    }
#else
#endif
    digitalWrite(SDCARD_CS,HIGH);//Deselect the SD card

    tNow = timelocal(now());
    sprintf_P(Dbuf,PSTR("%02d/%02d/%4d %02d:%02d:%02d"),month(tNow),day(tNow),year(tNow),hour(tNow),minute(tNow),second(tNow));
    Serial.println(Dbuf);
    new_minute = 0;
    sec_Timers[SETCLOCK_TIMER] = 60;
    sec_Timers[GETCLOCK_TIMER] = 60;
    sec_Timers[LCD_UPDATE_TIMER] = 59;
    sec_Timers[SEND_EMAIL_TIMER] = 0;
    min_Timers[X10_CHECK_TIMER] = 5 - (minute(tNow) % 5);
    min_Timers[NTP_RETRY_TIMER] = (24 * 60) + 10;
    min_Timers[EEPROM_SAVE_TIMER] = 5;
    min_Timers[EXIT_WOPEN_TIMER] = 0;
    sec_Timers[EXIT_WCLOSED_TIMER] = 0;
    msec_Timers[X10_ACTION_TIMER] = 1000;
    perform_x10_action = 0;
    action_ctr = 0;

// X10.init(rts, dtr, delay);
// RTS -> pin 7 on DB9
// DTR -> pin 4 on DB9
// GROUND pin 5 on DB9
//________________
//\ 1  2  3  4  5/
// \ 6  7  8  9 /
//  \----------/
    X10.init( X10RTS, X10DTR, 2);
    backup_unit = !digitalRead(SECOND_UNIT_SW);	// button pressed
    // all of this next bit is to set night mode on the development
    // unit to be 5 minutes from when it boots, lasting 5 minutes.
    if(backup_unit) {
	alarm_config.night_start[1] = hour(tNow);
	alarm_config.night_start[0] = minute(tNow) + 5;
	alarm_config.night_stop[1] = hour(tNow);
	alarm_config.night_stop[0] = minute(tNow) + 10;
	if(alarm_config.night_start[0] > 59) {
	    alarm_config.night_start[0] -= 60;
	    alarm_config.night_start[1]++;
	    if(alarm_config.night_start[1] > 23) {
		alarm_config.night_start[1] -= 24;
	    }
	}
	if(alarm_config.night_stop[0] > 59) {
	    alarm_config.night_stop[0] -= 60;
	    alarm_config.night_stop[1]++;
	    if(alarm_config.night_stop[1] > 23) {
		alarm_config.night_stop[1] -= 24;
	    }
	}
    }

    // get initial pot value to avoid spurious input character on startup
    first_poll_second_fp = 1;
    eeprom_restore_state();
    eeprom_restore_alm_idx();
#if DO_NOT_LOG_REBOOTS
    // backup log one notch, so state transition to restored state
    // will overwrite the last state transition
    alm_evt_idx--;
    alm_evt_idx %= MAX_AEVENTS;
#endif
    if(system_state < IDLE || system_state > NITEALRT) {
	system_state = IDLE;
    }
    quick_scan_loop();
    switch(system_state) {
	case IDLE :
	    disarm_system();
	    break;
	case WOPEN :
	    set_system_wait();
	    break;
	case WCLOSED :
	    set_ui_mode(UI_ARMED, 0);
	    set_ui_mode(UI_ARMED, 1);
	    set_system_wclosed();
	    break;
	case ARMED :
	    set_ui_mode(UI_ARMED, 0);
	    set_ui_mode(UI_ARMED, 1);
	    if(alarm_config.alarm_zone) {
		alert_zone = alarm_config.alarm_zone;
		alert_owner(quick_scan_loop() || alarm_config.alarm_zone);
	    }
	    else {
		arm_system();
	    }
	    break;
	case ALERT :
	    set_ui_mode(UI_ARMED, 0);
	    set_ui_mode(UI_ARMED, 1);
	    // in case door has closed since ALERT and REBOOT, get stored
	    // value of the zone that caused the alarm.
	    alert_owner(quick_scan_loop() || alarm_config.alarm_zone);
	    break;
	case NIGHT :
	    set_ui_mode(UI_NIGHT, 0);
	    set_ui_mode(UI_NIGHT, 1);
	    set_system_night();
	    break;
	case NITEALRT :
	    set_ui_mode(UI_NIGHT, 0);
	    set_ui_mode(UI_NIGHT, 1);
	    night_alert_owner(alarm_config.alarm_zone);
	    break;
    }
#if INTERNAL_WDT
    wdt_enable(WDTO_8S);
#else
    WDT_enable();
#endif
    button_E_pressed = 0;
}

unsigned int freeRam () {
extern int __heap_start, *__brkval;
int v;
    return (unsigned int) &v - (__brkval == 0 ? (int) &__heap_start : (int) __brkval);
}
// start of webclient.cpp
// #include "alarm.h"

const char alarmtmplstart[] PROGMEM = "<TABLE border=1>"
"<TR><TH colspan=4>Alarm</TH></TR>\n"
"<tr>\n"
"<td colspan=1 align=center>_DATE</td><td colspan=2 align=center>_TIME</td>\n"
"</tr>\n"
"<tr>\n"
"<td colspan=1 align=center>Status</td><td colspan=2 align=center>_SYSSTAT</td>\n"
"</tr>\n"
"<tr>\n"
"<td colspan=1 align=center>Battery</td><td colspan=2 align=center>_BATTV</td>\n"
"</tr>\n"
"<tr>\n"
"<td colspan=1 align=center>Uptime</td><td colspan=2 align=center>_UPTIME</td>\n"
"</tr>\n";

const char alarmrowtmpl[] PROGMEM = "<tr>\n"
"<th colspan=3 align=center>_NAME</th>\n"
"</tr>\n"
"<tr>\n"
"<td>Current Status:</td><td align=center>?AUMA</td><td>_STATUS</td>\n"
"</tr>\n"
"_EOF\n";

const char almlogtmplstart[] PROGMEM = "<TABLE border=1>"
"<TR><TH colspan=3>Alarm Log</TH></TR>\n"
"<tr>\n"
"<td align=center>Timestamp</td><td align=center>State</td><td align=center>open zones</td>\n"
"</tr>\n";

P(root_link) = "<A HREF=\"/\">Back to Top</A><br>";
P(alarm_link) = "<A HREF=\"/alarm\">Alarm Zones</A><br>";
P(almlog_link) = "<A HREF=\"/almlog\">Alarm Log</A><br>";
P(houselink) = "<A HREF=\"/housecodes\">House Codes</A><br>";
P(unitslink) = "<A HREF=\"/unit_codes\">Unit Assignments</A><br>";
P(schedlink) = "<A HREF=\"/schedule\">Schedule Table</A><br>";
P(manuallink) = "<A HREF=\"/manual\">Manual Control</A><br>";
P(configlink) = "<A HREF=\"/config\">Configuration</A><br>";
P(uptimestart) = "<br>uptime: ";
P(angle_br) = "<br>";
P(htmlHead) =
    "<html>"
    "<head>"
    "<title>Home Automation</title>"
    "<style type=\"text/css\">"
    "BODY { font-family: sans-serif }"
    "H1 { font-size: 14pt; text-decoration: underline }"
    "P  { font-size: 10pt; }"
    "</style>"
    "</head>"
    "<body>";
P(htmlTail) =
    "</body>"
    "</html>";
P(formend) = "<input type='submit' value='Submit'></form>";
P(houseformstart) = "<form action='" PREFIX "/housecodes' method='post'>\n";
P(unitformstart)  = "<form action='" PREFIX "/unit_codes' method='post'>\n";
P(eventformstart) = "<form action='" PREFIX "/schedule' method='post'>\n";
P(manualformstart) = "<form action='" PREFIX "/manual' method='post'>\n";
P(configformstart) = "<form action='" PREFIX "/config' method='post'>\n";
P(alarm_formstart) = "<form action='" PREFIX "/alarm' method='post'>\n";

// read a newline terminated line from the web template
PGM_P read_template_line (char *dbuf, PGM_P s) {
char c;
    *dbuf = '\0';
    while (c = pgm_read_byte(s)) {
	if(c == '\0') {
	    *dbuf++ = '\0';
	    return NULL;
	}      // leave pointing to '\0'
	else if(c == '\n') {
	    *dbuf++ = c;
	    *dbuf++ = '\0';
	    s++;         // point to next line or '\0'
	    break;
	}
	else {
	    *dbuf++ = c;
	    s++;
	}
    }
    return(s);
}

void show_template(WebServer &server, WebServer::ConnectionType type, bool addControls = false, unsigned char zone = 0, PGM_P ptmpl = alarmrowtmpl) {
char *p, *d;
unsigned char sp;
char Dbuf3[650];
char sname[12];
    do {
	ptmpl = read_template_line (Dbuf, ptmpl);
	for(p = Dbuf, d = Dbuf3; *p ; ) {
	    if(*p == '?') {
		*d = '\0';
		if(!strncmp_P(p, PSTR("?AUMA"), 5)) {
		    sprintf_P(d, alarm_config.zone_config[zone].enabled ? PSTR("<label><input type='radio' name='auma%d' value='e' checked>Enabled</label><label><input type='radio' name='auma%d' value='d'>Disabled</label>") : PSTR("<label><input type='radio' name='auma%d' value='e'>Enabled</label><label><input type='radio' name='auma%d' value='d' checked>Disabled</label>"), zone, zone);
		    d = Dbuf3 + strlen(Dbuf3);
		    p += 5;
		}
		else {
		    *d++ = *p++;
		}
	    }
	    else if(*p == '_') {
		*d = '\0';
		if(!strncmp_P(p, PSTR("_EOF"), 4)) {
		    ptmpl = NULL;
		    return;
		}
		else if(!strncmp_P(p, PSTR("_NAME"), 5)) {
		    strcat_P(d, (char*)pgm_read_word(&(zone_names[zone])));
		    d = Dbuf3 + strlen(Dbuf3);
		    p += 5;
		}
		else if(!strncmp_P(p, PSTR("_STATUS"), 7)) {
                    strcat(d, open[zone] ? "OPEN" : "closed");
		    d = Dbuf3 + strlen(Dbuf3);
		    p += 7;
		}
		else if(!strncmp_P(p, PSTR("_AUMA"), 5)) {
		    strcpy_P(d, alarm_config.zone_config[zone].enabled ? PSTR("Enabled") : PSTR("Disabled"));
		    d = Dbuf3 + strlen(Dbuf3);
		    p += 5;
		}
		else if(!strncmp_P(p, PSTR("_TIME"), 5)) {
		    sprintf_P(d, PSTR("%02d/%02d/%4d") ,month(tNow),day(tNow),year(tNow));
		    d += 10;
		    *d = '\0';
		    p += 5;
		}
		else if(!strncmp_P(p, PSTR("_SYSSTAT"), 8)) {
		    strncpy_P(sname, (char*)pgm_read_word(&(sstate_names[(int)system_state])), sizeof(sname));
		    strcpy(d, sname);
		    d += strlen(sname);
		    *d = '\0';
		    p += 8;
		}
		else if(!strncmp_P(p, PSTR("_BATTV"), 6)) {	// show battery voltage
		    unsigned int v = ((long)average_batt_reading * (long)config_data.batt_cal_constant) >> 10;
		    sprintf_P(d, PSTR("%2d.%02dV"), v/100, v%100);
		    d += 6;
		    *d = '\0';
		    p += 6;
		}
		else if(!strncmp_P(p, PSTR("_UPTIME"), 7)) {	// show uptime
		    unsigned int updays = uptime_min/1440L ;
		    unsigned int uphrs = (uptime_min % 1440L)/60 ;
		    unsigned int upmins = uptime_min % 60L ;
		    sprintf_P(d, PSTR("%4d days %02d:%02d"), updays, uphrs, upmins);
		    d += 15;
		    *d = '\0';
		    p += 7;
		}
		else if(!strncmp_P(p, PSTR("_DATE"), 5)) {
		    sprintf_P(d, PSTR("%02d:%02d:%02d") ,hour(tNow),minute(tNow),second(tNow));
		    d += 8;
		    *d = '\0';
		    p += 5;
		}
		else {
		    *d++ = *p++;
		}
	    }
	    else if(*p == '%') {
		if(*++p == 'd') {
		    sprintf(d, "%d", zone);
		    d++;
		    *d = '\0';
		    p++;
		}
		else {
		    *d++ = *p++;
		}
	    }
	    else if(*p == '\n') {
		*d++ = *p++;
		*p = '\0';
		*d = '\0';
		break;
	    }
	    else {
		*d++ = *p++;
	    }
	}
#if DEBUG_TMPL
	Serial.print(Dbuf);
	DLFn("^^^___Dbuf");
	Serial.print(Dbuf3);
	DLFn("^^^___Dbuf3");
#endif
	server.print(Dbuf3);
    } while(Dbuf[0] && ptmpl);
}

// show the alarm status/control page
void show_alarm_page(WebServer &server, WebServer::ConnectionType type, bool addControls = false) {
unsigned char zone;
char sname[20];
    if (addControls && system_state == IDLE) {
	server.httpSuccess("text/html");
    }
    else {
	server.httpSuccess("text/html","Refresh: 15\r\n");
    }
    server.printP(htmlHead);

    if (addControls && system_state == IDLE) server.printP(alarm_formstart);
    show_template(server, type, addControls, 0, alarmtmplstart);
    for(zone = 0; zone < MAXZONES ; zone++) {
	show_template(server, type, addControls, zone, alarmrowtmpl);
    }
    server.println(F("</table>"));
    if (addControls && system_state == IDLE) server.printP(formend);
    server.printP(root_link);
}

// show the alarm log page
// SEE ALSO show_alarm_log();
void show_almlog_page(WebServer &server, WebServer::ConnectionType type, bool addControls = false) {
unsigned int idx;
time_t gmtNow, lts;
AEVENT a;
unsigned int eeaddr;
char sname[10];
char zname[33];
    server.httpSuccess();
    server.printP(htmlHead);

    // if (addControls) server.printP(alarm_formstart);
    show_template(server, type, addControls, 0, almlogtmplstart);
    gmtNow = now();
    idx = alm_evt_idx;
    for(unsigned int j = 0 ; j < MAX_AEVENTS ; j++, --idx) {
	if(idx < 0) { idx = MAX_AEVENTS - 1; }
	eeaddr = ADDR_ALARM_LOG + (idx * sizeof(AEVENT));
	EEPROM_readAnything(eeaddr, a);
	if(a.ts > gmtNow) { continue; }
	if(a.ts < 1455000000L) { continue; }
	lts = timelocal(a.ts);
	sprintf_P(Dbuf,PSTR("%02d/%02d/%4d %02d:%02d:%02d  "),month(lts),day(lts),year(lts),hour(lts),minute(lts),second(lts));
	server.print(F("<tr><td>"));
	server.print(Dbuf);
	server.print(F("</td><td>"));
	strncpy_P(sname, (char*)pgm_read_word(&(sstate_names[(int)a.sstate])), sizeof(sname));
	server.println(sname);
	server.print(F("</td><td>"));
	if(a.openv) {
	    memset(zname, '\0', sizeof(zname));
	    for(uint8_t zone = 0 ; zone < MAXZONES ; zone++) {
		// DVFn("zone =", zone);
		if(a.openv & 1) {
		    // DVFn("active zone =", zone);
		    strncat_P(zname, (char*)pgm_read_word(&(short_zone_names[zone])), sizeof(zname) - strlen(zname));
		    strcat_P(zname, PSTR(" "));
		    // DVFn("zname =", zname);
		}
		a.openv >>= 1;
	    }
	    server.print(zname);
	}
	else {
	    server.print("none");
	}
	server.print(F("</td></tr>"));
    }
    server.println(F("</table>"));
    // if (addControls) server.printP(formend);
    server.printP(root_link);
}
// end of webclient.cpp

// returns 0 if none open, zone + 1 if where zone is first open found
uint8_t quick_scan_loop (void) {
uint8_t first = 0;
    open_count = 0;
    open_vector = 0;
    open_enabled_vector = 0;
    for(uint8_t zone = 0 ; zone < MAXZONES ; zone++) {
	if(digitalRead(alarm_config.zone_config[zone].contact) ^ INPUT_SEX) {
	    open_vector |= 1 << zone;
	    if(alarm_config.zone_config[zone].enabled) {
		open_enabled_vector |= 1 << zone;
	    }
	    open_count++;
	    last_open[zone] = open[zone] = DEBOUNCE;
	    if(!first) { first = zone + 1; }
	}
    }
    return first;
}

void scan_loop (uint8_t zone) {
uint8_t change = 0;
    // if contact is "open" (door is open, PIR is active, etc)
    if(digitalRead(alarm_config.zone_config[zone].contact) ^ INPUT_SEX) {
	if(open[zone] <= DEBOUNCE) {
	    open[zone]++;
	}
	if(open[zone] == DEBOUNCE) {
	    // DVF("zone:", zone);
	    // DVFn("debounce =", open[zone]);
	    if(!(open_vector & (1 << zone))) {
		open_count++;
		last_open[zone] = open[zone];
		change = 1;
	    }
	    open_vector |= 1 << zone;
	    DHF("updated (1) open_vector =", open_vector);
	    DVFn(", open_count =", open_count);
	    DLFn("set turn_on_backlight");
	    display[0].turn_on_backlight = 1;
	    display[1].turn_on_backlight = 1;
	    if(alarm_config.zone_config[zone].enabled) {
		open_enabled_vector |= 1 << zone;
		DHFn("updated (1) open_enabled_vector =", open_enabled_vector);
		if(change) {
		    char lzname[LONGEST_NAME];	// max length of any string in zone_names[]
		    strncpy_P(lzname, (char*)pgm_read_word(&(zone_names[zone])), sizeof(lzname));
		    DVF("Open:", lzname);
		    DVFn(", num =", zone);
		    HandleDisplay(0);
		    HandleDisplay(1);
		    if(system_state == WOPEN) {
			// if(zone == 0 || zone == 1)
			DLFn("In state WOPEN");
			if((alarm_config.zone_config[zone].attributes & (ATTR_NORMAL | ATTR_OUTSIDE)) == (ATTR_NORMAL | ATTR_OUTSIDE)) {
			    DLFn("Found NORMAL OUTSIDE door open --> WCLOSED");
			    set_system_wclosed();
			}
		    }
		    else if(system_state == ARMED) {
			DLFn("In state ARMED");
			if(!alarm_config.alarm_zone) {
			    alarm_config.alarm_zone = 1 + zone;
			    eeprom_save_alarm_config();
			}
			if(!is_entry_delay) {
			    DLFn("not in entry_delay");
			    release_the_hounds();	// play barking noise
			    inside_lights_on();	 // happens if it's night time
			}
			if(!sec_Timers[zone]) {	// if not in delay already
			    sec_Timers[zone] = alarm_config.zone_config[zone].entry_delay_sec;
			    // if(is_entry_delay < MAXZONES + 1)
			    if(!is_entry_delay)
			    {
				is_entry_delay = 1;
				DVFn("is_entry_delay =", is_entry_delay);
				DVFn("timer =", sec_Timers[zone]);
			    }
			}
		    }
		    else if(system_state == NIGHT) {
			if(!alarm_config.alarm_zone) {
			    alarm_config.alarm_zone = 1 + zone;
			    eeprom_save_alarm_config();
			}
			if(alarm_config.zone_config[zone].attributes & ATTR_OUTSIDE) {
			    night_alert_owner(zone + 1);
			}
		    }
		}
	    }
	}
    }
    else {
	if(open[zone]) {
	    open[zone]--;
	    if(!open[zone] && last_open[zone]) {
		char lzname[LONGEST_NAME];	// max length of any string in zone_names[]
		if((open_vector & (1 << zone))) {
		    open_vector &= ~(1 << zone);
		    open_enabled_vector &= ~(1 << zone);
		    if(open_count) { open_count--; }
		    last_open[zone] = open[zone];
		    change = 1;
		}
		strncpy_P(lzname, (char*)pgm_read_word(&(zone_names[zone])), sizeof(lzname));
		DVFn("just closed", lzname);
		// DHFn("AND mask:", ~(1 << zone));
		DLFn("set turn_on_backlight");
		display[0].turn_on_backlight = 1;
		display[1].turn_on_backlight = 1;
		DVF("updated (2) open_vector =", open_vector);
		DVFn(", open_count =", open_count);
		if(system_state == WCLOSED) {
		    DLF("newly closed and mode = WCLOSED");
		    // if(zone == 0 || zone == 1)
		    if((alarm_config.zone_config[zone].attributes & (ATTR_NORMAL | ATTR_OUTSIDE)) == (ATTR_NORMAL | ATTR_OUTSIDE)) {
			sec_Timers[EXIT_WCLOSED_TIMER] = 0;
			// hack to temporarily disable PIR sensor
			// if(alarm_config.zone_config[6].enabled) { }
			open_enabled_vector &= ~(1 << 6);
			if(arm_system()) {
			    DVFn("cannot arm, open zones (enabled_vector):", open_enabled_vector);
			    disarm_system();
			}
			else {
			    DLFn(" --> ARMED");
			}
		    }
		}
		else if(alarm_config.zone_config[zone].enabled) {
		    if(change) {
			char lzname[LONGEST_NAME];	// max length of any string in zone_names[]
			strncpy_P(lzname, (char*)pgm_read_word(&(zone_names[zone])), sizeof(lzname));
			HandleDisplay(0);
			HandleDisplay(1);
		    }
		}
	    }
	}
    }
    // find open zones that were enabled but are now disabled, and hide them
    if(open_enabled_vector) {
	for(uint8_t zone = 0 ; zone < MAXZONES ; zone++) {
	    if((open_enabled_vector & (1 << zone))) {
		if(!alarm_config.zone_config[zone].enabled) {
		    open_enabled_vector &= ~(1 << zone);
		}
	    }
	}
    }
    // find open zones that were disabled but are now enabled, and show them
    if(open_vector) {
	for(uint8_t zone = 0 ; zone < MAXZONES ; zone++) {
	    if((open_vector & (1 << zone))) {
		if(alarm_config.zone_config[zone].enabled) {
		    open_enabled_vector |= 1 << zone;
		}
	    }
	}
    }
    if(open_enabled_vector == 0 && system_state == IDLE) {
	is_entry_delay = 0;
	alarm_config.alarm_zone = 0;
    }
}

byte get_char (void) {
    while(!Serial1.available()) {
	;
    }
    return Serial1.read();
}

/*
mode: idle
show zones that are "unhappy"

"clear" button -> clears password attempt
"happy" && enter code -> mode: alarmed
"log" button -> mode: show log
"zone" button -> mode: ZSET (enable/disable zones)
---------------------------------------
mode: alarmed
unhappy zone -> wait delay -> raise the alarm
enter code -> mode: idle

---------------------------------------
mode: show log
navigate alarm history
"exit" button -> mode: idle

---------------------------------------
mode: zone set
pot controls which zone is displayed
"enter" key causes zone to toggle active/inactive
"clear" - exits to idle mode

---------------------------------------
*/
void inside_lights_on (void) {
int suntime;
uint8_t srh, srm, ssh, ssm, i;
int hhmm, shhmm, rhhmm;
    tNow = timelocal(now());
    suntime = sunrise(1, day(tNow), month(tNow), weekday(tNow));
    srh = suntime / 100;
    srm = suntime % 100;
    suntime = sunrise(0, day(tNow), month(tNow), weekday(tNow));
    ssh = suntime / 100;
    ssm = suntime % 100;
    hhmm = minute(tNow) + (hour(tNow) * 100);
    rhhmm = srm + (srh * 100);
    shhmm = ssm + (ssh * 100);
    if(hhmm > shhmm || hhmm < rhhmm) {	// if after sunset or before sunrise
	// turn on the inside lights
	for(i = 0 ; i < MAX_HOUSE_CODES * 16 ; i++) {
	    if(config_data.x10devices[i].is_inside_light) {
		action_queue[action_ctr].evidx = i;
		action_queue[action_ctr].action = 1;
		action_queue[action_ctr++].ctr = 4;
	    }
	}
    }
}

void night_alert_owner (uint8_t zonep1) {
    system_state = NITEALRT;
    eeprom_save_state();
    log_alarm();
    DLFn("system_state: NITEALRT");
    release_the_hounds();
    HandleDisplay(0);
    HandleDisplay(1);
    inside_lights_on();
}

void set_system_night (void) {
    system_state = NIGHT;
    eeprom_save_state();
    log_alarm();
    set_ui_mode(UI_NIGHT, 0);
    set_ui_mode(UI_NIGHT, 1);
    HandleDisplay(0);
    HandleDisplay(1);
}

void alert_owner (uint8_t zonep1) {
char lzname[LONGEST_NAME];	// max length of any string in zone_names[]
    system_state = ALERT;
    eeprom_save_state();
    if(!alert_zone) {
	alert_zone = alarm_config.alarm_zone;
    }
    log_alarm();
    DLFn("system_state: ALERT");
    alarm_bell_on();
    strncpy_P(lzname, (char*)pgm_read_word(&(zone_names[zonep1-1])), sizeof(lzname));
    DVFn("Alarm in zone", lzname);
    sec_Timers[SEND_EMAIL_TIMER] = 0;
    need_to_send_email = 1;
    inside_lights_on();
    HandleDisplay(0);
    HandleDisplay(1);
}

void set_system_wait (void) {
    system_state = WOPEN;
    eeprom_save_state();
    log_alarm();
    set_ui_mode(UI_ARMED, 0);
    set_ui_mode(UI_ARMED, 1);
    DVFn("starting WOPEN timer (min):", 10);
    // start 10 minute timer
    min_Timers[EXIT_WOPEN_TIMER] = 10;
    HandleDisplay(0);
    HandleDisplay(1);
}

void set_system_wclosed (void) {
    system_state = WCLOSED;
    eeprom_save_state();
    log_alarm();
    min_Timers[EXIT_WOPEN_TIMER] = 0;
    DVFn("stopping WOPEN timer, starting WCLOSED timer (sec):", 120);
    sec_Timers[EXIT_WCLOSED_TIMER] = 120;
    HandleDisplay(0);
    HandleDisplay(1);
}

// returns 0 for success, >0 for failure
uint8_t arm_system (void) {
uint8_t open_found = 0;
uint8_t zone;

    if(open_enabled_vector) { return(1); }
    system_state = ARMED;
    eeprom_save_state();
    log_alarm();
    alarm_bell_off();
    HandleDisplay(0);
    HandleDisplay(1);
    return(0);
}

void disarm_system (void) {
    for(uint8_t zone = 0 ; zone < MAXZONES ; zone++) {
	sec_Timers[zone] = 0;
    }
    system_state = IDLE;
    eeprom_save_state();
    sec_Timers[EXIT_WCLOSED_TIMER] = 0;
    min_Timers[EXIT_WOPEN_TIMER] = 0;
    sec_Timers[SEND_EMAIL_TIMER] = 0;
    need_to_send_email = 0;
    alarm_config.alarm_zone = 0;
    eeprom_save_alarm_config();
    set_ui_mode(UI_IDLE, 0);
    set_ui_mode(UI_IDLE, 1);
    log_alarm();
    corral_the_hounds();
    alarm_bell_off();
    HandleDisplay(0);
    HandleDisplay(1);
}

// A10,11,12,13 buttons A0 potentiometer
// rev 2: A10,11,12,13,A2 buttons, 6,7 quadrature pot
// uses vars last_quad_pos_result
char second_fp_poll(void) {
uint16_t pot_val, result, r2;
static uint8_t priors[5];
    if(!digitalRead(SECOND_E) && priors[4]) { priors[4] = 0; return 'E'; }
    if(digitalRead(SECOND_E) && !priors[4]) { priors[4] = 1; }
    if(!digitalRead(SECOND_D) && priors[3]) { priors[3] = 0; return 'D'; }
    if(digitalRead(SECOND_D) && !priors[3]) { priors[3] = 1; }
    if(!digitalRead(SECOND_C) && priors[2]) { priors[2] = 0; return 'C'; }
    if(digitalRead(SECOND_C) && !priors[2]) { priors[2] = 1; }
    if(!digitalRead(SECOND_B) && priors[1]) { priors[1] = 0; return 'B'; }
    if(digitalRead(SECOND_B) && !priors[1]) { priors[1] = 1; }
    if(!digitalRead(SECOND_A) && priors[0]) { priors[0] = 0; return 'A'; }
    if(digitalRead(SECOND_A) && !priors[0]) { priors[0] = 1; }
    result = 0x8000 | quad1.position();
    if(result != last_quad_pos_result) {
	last_quad_pos_result = result;
	// DVFn("q:", ((result & 0x7fff) + '0'));
	return((result & 0x7fff) + '0');
    }
    return '\0';
}

char second_fp_available (void) {
char tempc;
    tempc = second_fp_poll();
    if(tempc) {
      second_fp_char = tempc;
    }
    if(first_poll_second_fp) {
      first_poll_second_fp = 0;
      second_fp_char = 0;
    }
    return(second_fp_char != 0);
}

char second_fp_read (void) {
char tempc;
    tempc = second_fp_char;
    second_fp_char = '\0';
    return(tempc);
}

void HandlePanelUI (uint8_t di) {
char ch, ch2, reset_cnt;
uint8_t pwfound;
int kdir;
    if ((di ==  backup_unit && Serial1.available())
    ||  (di == !backup_unit && second_fp_available())) {
	uint8_t update_display = 0;
	if(di == backup_unit) {
	    ch = Serial1.read();
	}
	else {
	    ch = second_fp_read();
	    DVFn("HPUI: char =", ch);
	}
	// DVFn("set turn_on_backlight", di);
	display[di].turn_on_backlight = 1;
	switch(ch) {
	    case '0' : case '1' : case '2' : case '3' : case '4' :
	    case '5' : case '6' : case '7' : case '8' : case '9' :
		if(ui_mode[di] == UI_IDLE
		|| ui_mode[di] == UI_ARMED
		|| ui_mode[di] == UI_NIGHT) {
		    last_num[di] = ch;
		    passwd[di][pwp[di]] = last_num[di];
		    DVFn("passwd", passwd[di]);
		}
		else if(ui_mode[di] == UI_ZSET) {
		    last_num[di] = ch;
		    ch -= '0';
		    if(ch <= MAXZONES) {
			current_zone[di] = ch;
			DVFn("current_zone[di] =", current_zone[di]);
		    }
		}
		else if(ui_mode[di] == UI_LOG) {
		    // figure out knob direction
		    if(ch > last_num[di]) { uil_dir[di] = 0; }
		    if(ch < last_num[di]) { uil_dir[di] = 1; }
		    if(ch == '9') {
			last_num[di] = '1';
			if(di == 0) {
			    LCDquadpotPos(0, 1);
			}
			else {
			    quad1.position(1); // set knob to 1
			    last_quad_pos_result = 0x8001;
			}
		    }
		    else if(ch == '0') {
			last_num[di] = '8';
			if(di == 0) {
			    LCDquadpotPos(0, 8);
			}
			else {
			    quad1.position(8); // set knob to 8
			    last_quad_pos_result = 0x8008;
			}
		    }
		    else {
			last_num[di] = ch;
		    }
		    if(uil_dir[di]) {	// forwards in time
			uil_idx[di]++;
			uil_idx[di] %= MAX_AEVENTS;
			uil_cnt[di]--;
			if(uil_cnt[di] < 0) { uil_cnt[di] += MAX_AEVENTS; }
		    }
		    else {	// "normal" backwards in time
			uil_idx[di]--;
			if(uil_idx[di] < 0) { uil_idx[di] += MAX_AEVENTS; }
			uil_cnt[di]++;
			uil_cnt[di] %= MAX_AEVENTS;
		    }
		}
		update_display = 1;
		reset_cnt = 0;
		break;
	    case 'A' :	// "enter" key
		reset_cnt = 0;
		if(ui_mode[di] == UI_IDLE
		|| ui_mode[di] == UI_ARMED
		|| ui_mode[di] == UI_NIGHT) {
		    DVFn("mode is UI_IDLE, UI_ARMED or UI_NIGHT ?", ui_mode[di]);
		    DVFn("last_num[di]", last_num[di]);
		    pwfound = 0;
		    DVFn("passwd =", passwd[di]);
		    if(last_num[di] && pwp[di] < 10 ) {
			passwd[di][++pwp[di]] = '\0';
			DVFn("pwp =", pwp[di]);
			if(pwp[di] > 2) {
			    // check for valid password
			    for(uint8_t i = 0 ; i < 3 ; i++) {
				DVFn("stored =", alarm_config.password[i]);
				if(!strncmp(alarm_config.password[i], passwd[di], sizeof(passwd)/2)) {
				    DLFn("pw match");
				    pwfound++;
				    break;
				}
			    }
			    if(pwfound) {
				if(ui_mode[di] == UI_IDLE) {
				    if(open_enabled_vector) {
					for(char zone = 0 ; zone < MAXZONES ; zone++) {
					    if(open[zone]) {
						char lzname[LONGEST_NAME];	// max length of any string in zone_names[]
						strncpy_P(lzname, (char*)pgm_read_word(&(zone_names[zone])), sizeof(lzname));
						// can't ARM, have open door
						update_display = 1;
						DVFn("can't(k) because open zone:", lzname);
						return;
					    }
					}
				    }
				    else {
					set_system_wait();
				    }
				}
				else {
				    disarm_system();
				}
				pwp[di] = 0;
				memset((void *)passwd[di], '\0', sizeof(passwd)/2);
			    }
			    else {
				update_display = 1;
			    }
			}
			else {
			    update_display = 1;
			}
		    }
		    last_num[di] = '\0';
		}
		else if(ui_mode[di] == UI_ZSET) {
		    DLFn("enter in ZSET mode");
		    DVFn("last_num[di] =", last_num[di]);
		    if(last_num[di]) {
			alarm_config.zone_config[current_zone[di]-1].enabled ^= 1;
			update_display = 1;
		    }
		}
		else if(ui_mode[di] == UI_LOG) {
		    last_num[di] = '\0';
		    uil_dir[di] = 0;
		    uil_idx[di]--;
		    if(uil_idx[di] < 0) { uil_idx[di] = MAX_AEVENTS - 1; }
		    uil_cnt[di]++;
		    uil_cnt[di] %= MAX_AEVENTS;
		    DVFn("UI_LOG<+>:", uil_idx[di]);
		    update_display = 1;
		}
		break;
	    case 'B' :	// clear password or reset
		if(ui_mode[di] == UI_ZSET || ui_mode[di] == UI_LOG) {
		    set_ui_mode(UI_IDLE, di);
		}
		else {
		    pwp[di] = 0;
		    reset_cnt++;
		    if(reset_cnt == 3) {
			// do something.
			DLFn("can we get here?");
		    }
		    memset((void *)passwd[di], '\0', sizeof(passwd)/2);
		    last_num[di] = '\0';
		}
		update_display = 1;
		break;
	    case 'C' :
		if(ui_mode[di] == UI_LOG) {
		    last_num[di] = '\0';
		    uil_dir[di] = 1;
		    uil_idx[di]++;
		    uil_idx[di] %= MAX_AEVENTS;
		    uil_cnt[di]--;
		    if(uil_cnt[di] < 0) uil_cnt[di] += MAX_AEVENTS;
		    else                uil_cnt[di] %= MAX_AEVENTS;
		    DVFn("UI_LOG<->:", uil_idx[di]);
		    update_display = 1;
		}
		else if(ui_mode[di] == UI_ZSET) {
		    DLFn("C in ZSET mode");
		    DVFn("last_num[di] =", last_num[di]);
		    if(last_num[di]) {
			alarm_config.zone_config[current_zone[di]-1].enabled = 1;
			update_display = 1;
		    }
		}
		else if(ui_mode[di] == UI_IDLE) {
		    set_ui_mode(UI_ZSET, di);
		    last_num[di] = '\0';
		    reset_cnt = 0;
		}
		break;
	    case 'D' :
		if(ui_mode[di] == UI_LOG) {	// jump back 10 entries
		    last_num[di] = '\0';
		    uil_dir[di] = 0;
		    uil_idx[di] -= 10;
		    if(uil_idx[di] < 0) { uil_idx[di] += MAX_AEVENTS; }
		    uil_cnt[di] += 10;
		    uil_cnt[di] %= MAX_AEVENTS;
		    DVFn("UI_LOG<+10>:", uil_idx[di]);
		}
		else if(ui_mode[di] == UI_ZSET) {
		    DLFn("D in ZSET mode");
		    DVFn("last_num[di] =", last_num[di]);
		    if(last_num[di]) {
			alarm_config.zone_config[current_zone[di]-1].enabled = 0;
			update_display = 1;
		    }
		}
		else if(ui_mode[di] == UI_IDLE) {
		    set_ui_mode(UI_LOG, di);
		    last_num[di] = '\0';
		    reset_cnt = 0;
		    uil_dir[di] = 0;
		    uil_idx[di] = alm_evt_idx;
		    uil_cnt[di] = 0;
		    DVFn("new UI_LOG:", uil_idx[di]);
		}
		update_display = 1;
		break;
	    case 'E' :
		button_E_pressed = 2;
		update_display = 1;
		break;
	}
	if(update_display) {
	    HandleDisplay(di);
	}
    }
}

unsigned int fram;
unsigned int lmsgcnt;

void poll_disp_width (void) {
static unsigned char old20[2];
unsigned char new20;
    for(uint8_t di = 0 ; di < 2 ; di++) {
	new20 = (di) ? digitalRead(SEC_COL16SW) : digitalRead(PRI_COL16SW);
	config_data.display[di].cols = new20 ? 20 : 16;
	if(old20[di] != new20) {
	    DVFn("cols=", config_data.display[di].cols);
	}
	old20[di] = new20;
    }
}

void pollWireless (void) {
    if(!digitalRead(Wireless1)) {
	if(!wireless_state[0]) {
	    DLFn("Wireless 1 ON");
	    wireless_state[0]++;
	    DLFn("set turn_on_backlight");
	    display[0].turn_on_backlight = 1;
	    display[1].turn_on_backlight = 1;
	    if(ui_mode[0] == UI_IDLE) {
		if(open_enabled_vector) {
		    for(char zone = 0 ; zone < MAXZONES ; zone++) {
			if(open[zone]) {
			    char lzname[LONGEST_NAME];	// max length of any string in zone_names[]
			    strncpy_P(lzname, (char*)pgm_read_word(&(zone_names[zone])), sizeof(lzname));
			    // can't ARM, have open door
			    DVFn("can't(w) because open zone:", lzname);
			}
		    }
		    // play sad sound
		    play_sound(SADSOUND);
		}
		else {
		    // play happy sound
		    play_sound(TADAAA);
		    set_system_wait();
		    DLFn("going to WOPEN");
		}
	    }
	    // second button press arms the system
	    else if(system_state == WOPEN) {
		DLFn("two presses, going to ARMED");
		// play happy sound
		mydelay(10);
		play_sound(TADAAA);
		arm_system();
		mydelay(10);
		play_sound(TADAAA);
	    }
	}
    }
    else {
	if(wireless_state[0]) {
	    DLFn("Wireless 1 OFF");
	    wireless_state[0] = 0;
	}
    }
    if(!digitalRead(Wireless2)) {
	if(!wireless_state[1]) {
	    DLFn("Wireless 2 ON");
	    wireless_state[1]++;
	    DLFn("set turn_on_backlight");
	    display[0].turn_on_backlight = 1;
	    display[1].turn_on_backlight = 1;
	    sec_Timers[EXIT_WCLOSED_TIMER] = 0;
	    min_Timers[EXIT_WOPEN_TIMER] = 0;
	    if(system_state != IDLE) {
		disarm_system();
		// play other happy sound
		play_sound(NOTES);
	    }
	}
    }
    else {
	if(wireless_state[1]) {
	    DLFn("Wireless 2 OFF");
	    wireless_state[1] = 0;
	}
    }
}

void my_yield (void) {
    pollWireless();
    HandlePanelUI(0);	// no-op if no character waiting
    HandlePanelUI(1);	// no-op if no character waiting
#if INTERNAL_WDT
    // wdt_reset();
#else
    WDT_reset();
#endif
}

uint8_t try_set_backup_rtc (void) {
uint8_t set_clock_ok = 0;
uint8_t clock_tries = 0;
    do {
	time_t utcNow = now();	// get UTC time
	if(year(utcNow) >= 2016) {
	    sprintf_P(Dbuf,PSTR("rtc.set: %02d/%02d/%4d %02d:%02d:%02d UTC"),month(utcNow),day(utcNow),year(utcNow),hour(utcNow),minute(utcNow),second(utcNow));
	    Serial.println(Dbuf);
	    uint8_t res = rtc.set(utcNow);
	    time_t t = rtc.get();
	    if(res || year(t) != year(utcNow) || month(t) != month(utcNow)) {
		DVF("after rtc.set: bad year=", year(t));
		DVF(" or bad month=", month(t));
		DVFn(" or bad return value=", res);
		DVFn("clock_tries =", clock_tries);
		mydelay(500 * (1 + clock_tries));
	    }
	    else {
		set_clock_ok = 1;
		DLFn("rtc.set() and rtc.get() OK");
	    }
	}
    } while((++clock_tries < 8) && !set_clock_ok);
    config_data.clock_is_set = set_clock_ok;
    return(set_clock_ok);
}

int firsttime = true;
void loop (void) {
int index = 0;
int change;
uint8_t evctr;
static unsigned char zone;

#if INTERNAL_WDT
    // wdt_reset();
#else
    WDT_reset();
#endif
    pollWireless();
    if(alert_zone && alert_zone <= MAXZONES && alarm_config.alarm_zone) {
	// then zone alarm timed out
	alert_owner(alert_zone);
	email_alert_zone = alert_zone;
	alert_zone = 0;
    }
    if(need_to_send_email && !sec_Timers[SEND_EMAIL_TIMER] && email_alert_zone <= MAXZONES) {
	if(sendEmail(email_alert_zone - 1)) {
	    Serial.println(F("Email sent via Ethernet"));
	    need_to_send_email = 0;
	}
	else if(sendEmailWiFi(email_alert_zone - 1)) {
	    Serial.println(F("Email sent via WiFi"));
	    need_to_send_email = 0;
	}
	else {
	    wifi_setup();	// restart WiFi connection
	    if(sendEmailWiFi(email_alert_zone - 1)) {
		Serial.println(F("Email sent via WiFi"));
		need_to_send_email = 0;
	    }
	    else {
		Serial.println(F("Email failed"));
		email_failed_counter++;
		if(email_failed_counter < MAX_EMAIL_FAILS) {
		    sec_Timers[SEND_EMAIL_TIMER] = 3;
		}
		else {
		    resetFunc();	// try reboot, we're hosed
		}
	    }
	}
    }
    // do a keyboard scan every time through loop()
    // if dns_working, light LED.  button press inverts LED state so we
    // can test button operation.
    digitalWrite(DNS_OK_LED, dns_working ^ !digitalRead(USE_DHCP_BUTTON));
    PF(9);
    zone++;
    zone %= MAXZONES;
    poll_disp_width();
    HandlePanelUI(0);	// no-op if no character waiting
    HandlePanelUI(1);	// no-op if no character waiting
    
    if(msec_Timers[SOUND_TMR] == 0 && sound_needs_mute) {
	play_sound(MUTE);
	sound_needs_mute = 0;
    }
    if(msec_Timers[ASCAN_TMR] == 0) {
	msec_Timers[ASCAN_TMR] = ASCAN_DEL;
	scan_loop(zone);
    }
    if(firsttime == true) {
	msec_Timers[KBUITIMER] = 0;
	msec_Timers[TOD_TIMER] = 0;
    }
    if(msec_Timers[TOD_TIMER] == 0) {
	msec_Timers[TOD_TIMER] = 1000;
	tNow = timelocal(now());
    }
    for(byte i = 0 ; i < MAXDISPLAYS ; i++) {
	if(display[i].turn_off_backlight) {
	    PF(12);
	    display[i].turn_off_backlight = 0;
	    display[i].turn_on_backlight = 0;
	    LCDbacklight(i, (unsigned int)config_data.dimbacklight);
	}
	else if(display[i].turn_on_backlight) {
	    display[i].turn_on_backlight = 0;
	    display[i].backlight_timer = BLTIMEOUT;
	    LCDbacklight(i, (unsigned int)config_data.backlight);
	}
    }

    PF(16);
    if(!sec_Timers[SETCLOCK_TIMER] && dns_working) {
	if(!config_data.clock_is_set) {
	    DLFn("clock isn't set");
	    if(!try_set_backup_rtc()) {
		sec_Timers[SETCLOCK_TIMER] = 3600;	// try again in an hour
	    }
	    else {
		sec_Timers[SETCLOCK_TIMER] = 24 * 3600;	// check again in a day
	    }
	}
	else {
	    time_t nUTCnow, bUTCnow;
	    nUTCnow = now();
	    bUTCnow = rtc.get();
	    if((year(nUTCnow) != year(bUTCnow))
	    || (month(nUTCnow) != month(bUTCnow))
	    || (day(nUTCnow) != day(bUTCnow))) {
		DLFn("clock set, wrong time");
		if(year(nUTCnow) != year(bUTCnow)) {
		    DVF("NTPy =", year(nUTCnow));
		    DVFn(", DSy =", year(bUTCnow));
		}
		if(month(nUTCnow) != month(bUTCnow)) {
		    DVF("NTPm =", month(nUTCnow));
		    DVFn(", DSm =", month(bUTCnow));
		}
		if(day(nUTCnow) != day(bUTCnow)) {
		    DVF("NTPd =", day(nUTCnow));
		    DVFn(", DSd =", day(bUTCnow));
		}
		config_data.clock_is_set = 0;
		if(!try_set_backup_rtc()) {
		    sec_Timers[SETCLOCK_TIMER] = 3600;	// try again in an hour
		}
		else {
		    sec_Timers[SETCLOCK_TIMER] = 24 * 3600;	// check again in a day
		}
	    }
	}
    }
    if(system_state == WCLOSED && !sec_Timers[EXIT_WCLOSED_TIMER]) {
	DLFn("Exit WCLOSED Delay timeout, System -> IDLE");
	disarm_system();
    }

    PF(17);
#if SDCARD
    if(second() % 60 == 0 && minute() % 15 == 0 && !datalog_done[zone]) { //check if time for datalog
	datalog_done[zone]++;
	SDdatalog(zone);
    }
    if(second() % 60 == 2 && minute() % 15 == 0) {
	datalog_done[zone] = 0;
    }
#endif
    if(0 && zone == 0 && second() % 60 == 0 && minute() == 11) { //once/hour reset ethernet
	// pinMode(SDCARD_CS,OUTPUT);
	// digitalWrite(SDCARD_CS,HIGH);//Deselect the SD card
	// DL("EthReset");
	// ethernetReset();
#if EthHardCoded
	// Ethernet.begin(config_data.mac, config_data.ip, dns_server, gateway_ip, subnet_mask);
#else
	// Ethernet.begin(config_data.mac);
#endif
	// mydelay(1000);  // wait a bit for the card to stabilize
    }
    PF(18);

    //web server control code
    PF(19);
#if PROFILING
    if(second() % 60 == 3 && !prof_has_dumped) {
	dump_profiling_data();	// also clears profiling data
    }
    if(second() % 60 == 4 && prof_has_dumped) {
	prof_has_dumped = 0;
    }
#endif
    if(perform_x10_action && action_ctr) {
	perform_x10_action = 0;
	// ACTION action_queue[MAX_X10EVENTS]; evidx; action; ctr;
	uint8_t evidx = action_queue[action_ctr-1].evidx;
	uint8_t evhouse  = config_data.house_codes[evidx >> 4];
	uint8_t evunit   = evidx & 0x0f;
	send_command(evhouse, evunit, action_queue[action_ctr-1].action);
	if(action_queue[action_ctr-1].ctr) action_queue[action_ctr-1].ctr--;
	if(!action_queue[action_ctr-1].ctr) {
	    action_ctr--;
	}
    }
    if(new_minute) {
	new_minute = 0;
	if(min_Timers[EEPROM_SAVE_TIMER]) {
	    min_Timers[EEPROM_SAVE_TIMER]--;
	    if(!min_Timers[EEPROM_SAVE_TIMER]) {
		eeprom_save_all();
		min_Timers[EEPROM_SAVE_TIMER] = 5;
	    }
	}
	if(min_Timers[NTP_RETRY_TIMER]) {
	    min_Timers[NTP_RETRY_TIMER]--;
	    if(!min_Timers[NTP_RETRY_TIMER]) {
		min_Timers[NTP_RETRY_TIMER] = (24 * 60) + 10;
		// retry the dns -> ntp business again
		uint16_t dns_delay = 50;
		for(uint8_t retries = 0 ; retries < 3 ; retries++) {
		    DNSClient dns;
		    IPAddress remote_addr;
		    int ret;
		    DLFn("before dns.begin()");
		    dns.begin(Ethernet.dnsServerIP());
		    DLFn("before dns.getHostByName()");
		    ret = dns.getHostByName("pool.ntp.org", remote_addr);
		    if (ret == 1) {
			DVFn("address of pool.ntp.org", remote_addr);
			dns_working = 1;
			break;
		    }
		    else {
			DLF("(2) dns failed");
			DVFn(", retries =", retries);
			delay(dns_delay);
#if INTERNAL_WDT
			// wdt_reset();
#else
			WDT_reset();
#endif
			dns_delay <<= 1;
			dns_working = 0;
		    }
		}
	    }
	    if(ntp_fail_counter > MAX_NTP_FAILURES) {
		// internet is well and truly hosed
		resetFunc();
	    }
	}
	if(min_Timers[X10_CHECK_TIMER]) {
	    min_Timers[X10_CHECK_TIMER]--;
	    if(min_Timers[X10_CHECK_TIMER] == 0L) {	// check X10 schedule
		int suntime;
		uint8_t srh, srm, ssh, ssm, i;
		tNow = timelocal(now());
		sprintf_P(Dbuf,PSTR("X10: %02d/%02d/%4d %02d:%02d:%02d "),month(tNow),day(tNow),year(tNow),hour(tNow),minute(tNow),second(tNow));
		Serial.print(Dbuf);
		suntime = sunrise(1, day(tNow), month(tNow), weekday(tNow));
		srh = suntime / 100;
		srm = suntime % 100;
		sprintf_P(Dbuf, PSTR("sunrise: %d:%02d; "), srh, srm);
		Serial.print(Dbuf);
		suntime = sunrise(0, day(tNow), month(tNow), weekday(tNow));
		ssh = suntime / 100;
		ssm = suntime % 100;
		sprintf_P(Dbuf, PSTR("sunset: %d:%02d"), ssh, ssm);
		Serial.println(Dbuf);
		min_Timers[X10_CHECK_TIMER] = 5 - (minute(tNow) % 5);
		for(evctr = 0 ; evctr < MAX_X10EVENTS ; evctr++) {
		    if(config_data.x10events[evctr].action == 2) { continue; }
		    uint8_t evtime   = config_data.x10events[evctr].hour;
		    uint8_t evidx    = config_data.x10events[evctr].uidx;
		    uint8_t evaction = config_data.x10events[evctr].action;
#if ANNOUNCE_X10
		    sprintf_P(Dbuf, PSTR("%2d: %2d:00 %s %d"), evctr, evtime, config_data.x10devices[evidx].name, evaction);
		    Serial.println(Dbuf);
#endif
		    uint8_t smin = 0;
		    int hhmm, shhmm;
		    uint8_t debug = 1;
		    if(evtime == 32 || evtime == 33) {
			if(evtime == 33) {
			    evtime = ssh;
			    smin = ssm;
			}
			else {
			    evtime = srh;
			    smin = srm;
			}
			hhmm = minute(tNow) + (hour(tNow) * 100);
			shhmm = smin + (evtime * 100);
#if ANNOUNCE_X10
			if(debug) {
			    sprintf_P(Dbuf, PSTR("hhmm = %d, sun hhmm = %d"), hhmm, shhmm);
			    Serial.println(Dbuf);
			}
#endif
			if(abs(hhmm - shhmm) < 15) {
			    // ACTION action_queue[MAX_X10EVENTS]; uidx; action; ctr;
			    action_queue[action_ctr].evidx = evidx;
			    action_queue[action_ctr].action = evaction;
			    action_queue[action_ctr++].ctr = 20;
			    sprintf_P(Dbuf, PSTR("adding %d %d %d to action_queue[%d]"), evidx, evaction, 20, action_ctr-1);
			    Serial.println(Dbuf);
			}
		    }
		    else if(evtime == hour(tNow) && minute(tNow) < 5) {
			// ACTION action_queue[MAX_X10EVENTS]; uidx; action; ctr;
			action_queue[action_ctr].evidx = evidx;
			action_queue[action_ctr].action = evaction;
			action_queue[action_ctr++].ctr = 20;
			sprintf_P(Dbuf, PSTR("adding %d %d %d to action_queue[%d]"), evidx, evaction, 20, action_ctr-1);
			Serial.println(Dbuf);
		    }
		}
	    }
	}
	if(min_Timers[EXIT_WOPEN_TIMER]) {
	    min_Timers[EXIT_WOPEN_TIMER]--;
	    if(!min_Timers[EXIT_WOPEN_TIMER]) {
		if(open_enabled_vector) {
		    DLFn("Exit WOPEN Delay timeout, open door so System -> IDLE");
		    // play sad sound
		    play_sound(SADSOUND);
		    disarm_system();
		}
		else {
		    DLFn("Exit WOPEN Delay timeout, System -> ARMED");
		    // play happy sound
		    play_sound(TADAAA);
		    arm_system();
		}
	    }
	}
	if(system_state == IDLE
	&& alarm_config.night_mode_enabled
	&& alarm_config.night_start[1] == hour(tNow)
	&& alarm_config.night_start[0] == minute(tNow)) {
	    set_system_night();
	}
	if(system_state == NIGHT
	&& alarm_config.night_stop[1] == hour(tNow)
	&& alarm_config.night_stop[0] == minute(tNow)) {
	    disarm_system();
	}
    }
    if(lcd_needs_update) {
	HandleDisplay(0);
	HandleDisplay(1);
    }

    if(1) {
	WebServer webserver(PREFIX, 80);
	webserver.begin();

	webserver.setDefaultCommand(&defaultCmd);
	webserver.addCommand("housecodes", &houseCmd);
	webserver.addCommand("unit_codes", &unitCmd);
	webserver.addCommand("schedule",   &schedCmd);
	webserver.addCommand("manual",     &manualCmd);
	webserver.addCommand("config",     &configCmd);
	webserver.addCommand("alarm",      &alarm_Cmd);
	webserver.addCommand("almlog",     &almlog_Cmd);

	webserver.processConnection();
    }
    average_batt_filter += analogRead(A8) - average_batt_reading;
    average_batt_reading = average_batt_filter >> 4;

    if(firsttime == true) {
	fram = freeRam();
	DVFn("fram=", fram);
	lmsgcnt = 0;
	firsttime = false;
    }
    if(lmsgcnt) { lmsgcnt--; }
    if(!lmsgcnt && (fram != freeRam())) {
	DVFn("mem leak: Mem=", freeRam());
	lmsgcnt = 100;
    }

}

#define DEBUG 0

time_t timelocal(time_t utctime) {
time_t rt;
    if(config_data.tzindex == Eastern) {
	rt = usEastern.toLocal(utctime);
    }
    else if(config_data.tzindex == Central) {
	rt = usCentral.toLocal(utctime);
    }
    else if(config_data.tzindex == Mountain) {
	rt = usMountain.toLocal(utctime);
    }
    else if(config_data.tzindex == Arizona) {
	rt = usArizona.toLocal(utctime);
    }
    else if(config_data.tzindex == Pacific) {
	rt = usPacific.toLocal(utctime);
    }
    return(rt);
}


// commands are functions that get called by the webserver framework
// they can read any posted data from client, and they output to server
// The form is acted upon by configCmd()

void outputConfig(WebServer &server, WebServer::ConnectionType type, bool addControls = false) {
int i, h;
char sname[20], ename[4];
    server.httpSuccess();
    server.printP(htmlHead);

    if (addControls && system_state == IDLE) {
	server.printP(configformstart);
    }

  P(config_pagestart) =
      "<TABLE border=1>"
      "<TR><TH colspan=2>Configuration</TH></TR>\n";
    server.printP(config_pagestart);
    server.printP(PSTR("<TR><TD>Timezone: </TD>\n<TD>"));
    // radio buttions Eastern, Central, . . . Pacific
    if (addControls && system_state == IDLE) {
	char unit[5];
	strcpy_P(unit, PSTR("TZ"));
	server.radioButton(unit, "0", "Eastern", (config_data.tzindex == 0));
	server.radioButton(unit, "1", "Central", (config_data.tzindex == 1));
	server.radioButton(unit, "2", "Mountain", (config_data.tzindex == 2));
	server.radioButton(unit, "3", "Arizona", (config_data.tzindex == 3));
	server.radioButton(unit, "4", "Pacific", (config_data.tzindex == 4));
    }
    else {
	uint8_t tzidx;
	tzidx = (uint8_t)config_data.tzindex;
	server.printP((tzidx == 0) ? PSTR("Eastern") :
		      (tzidx == 1) ? PSTR("Central") :
		      (tzidx == 2) ? PSTR("Mountain") :
		      (tzidx == 3) ? PSTR("Arizona") : PSTR("Pacific"));
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>Battery A/D: </TD>\n<TD>"));
    unsigned int v = ((long)average_batt_reading * (long)config_data.batt_cal_constant) >> 10;
    sprintf_P(sname, PSTR("%d%02d0"), v/100, v%100);
    server.printP(PSTR("calib const: "));
    server << config_data.batt_cal_constant;
    server.printP(PSTR(" -> batt V : "));
    if (addControls && system_state == IDLE) {
	server.printP(PSTR("<input type=\"text\" size=\"5\" maxlength=\"5\" name=\"battvolts\" value=\""));
	server << sname;
	server.printP(PSTR("\">"));
    }
    else {
	server << sname;
    }
    server.printP(PSTR("&nbsp;mV"));
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>MAC address:</TD>\n<TD>"));
    for(i = 0 ; i < 6 ; i++) {
	if (addControls && system_state == IDLE) {
	    char fieldname[4];
	    fieldname[0] = 'm';
	    itoa(i, fieldname + 1, 10);
	    server.printP(PSTR("<input type=\"text\" size=\"3\" maxlength=\"3\" name=\""));
	    server << fieldname;
	    server.printP(PSTR("\" value=\""));
	    server.print(config_data.mac[i], HEX);
	    server.printP(PSTR("\">&nbsp;"));
	}
	else {
	    server.print(config_data.mac[i], HEX);
	    if(i < 5) { server << ' '; }
	}
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>my IP address:</TD>\n<TD>"));
    for(i = 0 ; i < 4 ; i++) {
	if (addControls && system_state == IDLE) {
	    char fieldname[4];
	    fieldname[0] = 'i';
	    itoa(i, fieldname + 1, 10);
	    server.printP(PSTR("<input type=\"text\" size=\"4\" maxlength=\"4\" name=\""));
	    server << fieldname;
	    server.printP(PSTR("\" value=\""));
	    server << config_data.ip[i];
	    server.printP(PSTR("\">&nbsp;"));
	}
	else {
	    server << config_data.ip[i];
	    if(i < 3) { server << '.'; }
	}
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>gateway IP address:</TD>\n<TD>"));
    for(i = 0 ; i < 4 ; i++) {
	if (addControls && system_state == IDLE) {
	    char fieldname[4];
	    fieldname[0] = 'g';
	    itoa(i, fieldname + 1, 10);
	    server.printP(PSTR("<input type=\"text\" size=\"4\" maxlength=\"4\" name=\""));
	    server << fieldname;
	    server.printP(PSTR("\" value=\""));
	    server << config_data.gateway_ip[i];
	    server.printP(PSTR("\">&nbsp;"));
	}
	else {
	    server << config_data.gateway_ip[i];
	    if(i < 3) { server << '.'; }
	}
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>dns IP address:</TD>\n<TD>"));
    for(i = 0 ; i < 4 ; i++) {
	if (addControls && system_state == IDLE) {
	    char fieldname[4];
	    fieldname[0] = 'd';
	    itoa(i, fieldname + 1, 10);
	    server.printP(PSTR("<input type=\"text\" size=\"4\" maxlength=\"4\" name=\""));
	    server << fieldname;
	    server.printP(PSTR("\" value=\""));
	    server << config_data.dns_server[i];
	    server.printP(PSTR("\">&nbsp;"));
	}
	else {
	    server << config_data.dns_server[i];
	    if(i < 3) { server << '.'; }
	}
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>subnet mask:</TD>\n<TD>"));
    for(i = 0 ; i < 4 ; i++) {
	if (addControls && system_state == IDLE) {
	    char fieldname[4];
	    fieldname[0] = 's';
	    itoa(i, fieldname + 1, 10);
	    server.printP(PSTR("<input type=\"text\" size=\"4\" maxlength=\"4\" name=\""));
	    server << fieldname;
	    server.printP(PSTR("\" value=\""));
	    server << config_data.subnet_mask[i];
	    server.printP(PSTR("\">&nbsp;"));
	}
	else {
	    server << config_data.subnet_mask[i];
	    if(i < 3) { server << '.'; }
	}
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>wifi credentials:</TD>\n<TD>"));
    if (addControls && system_state == IDLE) {
	// FOR DEBUGGING, close SECOND_UNIT_SW to see values of ssid & pass
	// otherwise, don't show this as it exposes the wifi credentials
	// to anyone.
	server.printP(PSTR("SSID: <input type=\"text\" size=\"10\" maxlength=\"30\" name=\"ssid\" value=\""));
	if(!digitalRead(SECOND_UNIT_SW)) {	// if low, show ssid & pass
	    server << config_data.wifi_ssid;
	}
	else {
	    server.printP(PSTR("********"));
	}
	server.printP(PSTR("\">&nbsp;"));
    }
    else {
	server.printP(PSTR("SSID: "));
	if(!digitalRead(SECOND_UNIT_SW)) {	// if low, show ssid & pass
	    server << config_data.wifi_ssid;
	}
	else {
	    server.printP(PSTR("********"));
	}
    }
    if (addControls && system_state == IDLE) {
	// FOR DEBUGGING, close SECOND_UNIT_SW to see values of ssid & pass
	// otherwise, don't show this as it exposes the wifi credentials
	// to anyone.
	server.printP(PSTR("password: <input type=\"text\" size=\"10\" maxlength=\"30\" name=\"pass\" value=\""));
	if(!digitalRead(SECOND_UNIT_SW)) {	// if low, show ssid & pass
	    server << config_data.wifi_pass;
	}
	else {
	    server.printP(PSTR("********"));
	}
	server.printP(PSTR("\">&nbsp;"));
    }
    else {
	server.printP(PSTR("&nbsp;password: "));
	if(!digitalRead(SECOND_UNIT_SW)) {	// if low, show ssid & pass
	    server << config_data.wifi_pass;
	}
	else {
	    server.printP(PSTR("********"));
	}
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("<TR><TD>night mode:</TD>\n<TD>"));
    sprintf(sname, "%02d", alarm_config.night_start[0]);
    sprintf(ename, "%02d", alarm_config.night_stop[0]);
    if (addControls && system_state == IDLE) {
	server.printP(alarm_config.night_mode_enabled ? PSTR("<label><input type='radio' name='nmen' value='1' checked>Enabled</label><label><input type='radio' name='nmen' value='0'>Disabled</label>") : PSTR("<label><input type='radio' name='nmen' value='1'>Enabled</label><label><input type='radio' name='nmen' value='0' checked>Disabled</label>"));
	server.printP(PSTR("&nbsp;&nbsp;start @ <input type=\"text\" size=\"2\" maxlength=\"2\" name=\"nmsh\" value=\""));
	server.print(alarm_config.night_start[1]);
	server.printP(PSTR("\">:"));
	server.printP(PSTR("<input type=\"text\" size=\"2\" maxlength=\"2\" name=\"nmsm\" value=\""));
	server.print(sname);
	server.printP(PSTR("\">&nbsp;"));
	server.printP(PSTR(" end @ <input type=\"text\" size=\"2\" maxlength=\"2\" name=\"nmeh\" value=\""));
	server.print(alarm_config.night_stop[1]);
	server.printP(PSTR("\">:"));
	server.printP(PSTR("<input type=\"text\" size=\"2\" maxlength=\"2\" name=\"nmem\" value=\""));
	server.print(ename);
	server.printP(PSTR("\">"));
    }
    else {
	server.printP(alarm_config.night_mode_enabled ? PSTR("Enabled") : PSTR("Disabled"));
	server.printP(PSTR("&nbsp;&nbsp;start \@"));
	server.print(alarm_config.night_start[1]);
	server.printP(PSTR(":"));
	server.print(sname);
	server.printP(PSTR(",&nbsp;end \@"));
	server.print(alarm_config.night_stop[1]);
	server.printP(PSTR(":"));
	server.print(ename);
    }
    server.printP(PSTR("</TD></TR>\n"));
    server.printP(PSTR("</table><br>"));
    server.printP(PSTR("To perform Battery A/D calibration, enter actual battery voltage in mV and click Submit<br><br>"));

    if (addControls && system_state == IDLE) server.printP(formend);
    server.printP(root_link);

    server.printP(htmlTail);
#if WEB_OUT_DEBUG
    DLFn("end outputConfig()");
#endif
}

void outputManual(WebServer &server, WebServer::ConnectionType type, bool addControls = false) {
    server.httpSuccess();
    server.printP(htmlHead);

    if (addControls && system_state == IDLE) {
	server.printP(manualformstart);
    }

    P(live_pagestart) =
      "<TABLE border=1>"
      "<TR><TH colspan=2>Manual Controls</TH></TR>\n";
    server.printP(live_pagestart);

    for (uint8_t h = 0; h <= 1; ++h) {	// both house codes
	for (uint8_t i = 0; i <= 15; ++i) {	// all 16 units per house code
	    char unit[5];
	    // hour uidx action
	    // config_data.x10events[i].hour
	    // config_data.x10events[i].uidx
	    // config_data.x10events[i].action
	    server.printP(PSTR("<TR><TD>"));
	    if(config_data.x10devices[(h * 16) + i].name[0]) {
		server << config_data.x10devices[(h * 16) + i].name ;
	    }
	    else {
		char housevalue[2];
		char fieldname[3];
		housevalue[0] = 'A' + config_data.house_codes[h];
		housevalue[1] = '\0';
		itoa((h*16)+i, &fieldname[0], 10);
		server << housevalue << fieldname ;
	    }
	    server << " &nbsp;";
	    server.printP(PSTR("</TD>\n<TD>"));
	    // radio buttions ON, OFF, NOOP
	    unit[0] = 'L';
	    itoa((h * 16) + i, unit + 1, 10);
	    server.radioButton(unit, "1", "On", 0);
	    server.radioButton(unit, "0", "Off", 0);
	    server.radioButton(unit, "2", "No-op", 1);
	    server.printP(PSTR("</TD></TR>\n"));
	}
    }
    server.printP(PSTR("</table><p>"));

    if (addControls && system_state == IDLE) server.printP(formend);
    server.printP(root_link);

    server.printP(htmlTail);
#if WEB_OUT_DEBUG
    DLFn("end outputManual()");
#endif
}

#define DEBUG_OUTP_EV 0
void outputEvents(WebServer &server, WebServer::ConnectionType type, bool addControls = false) {
uint8_t i;
#if WEB_OUT_DEBUG
// char sbuf[50];
#endif
    server.httpSuccess();
    server.printP(htmlHead);

    if (addControls) {
	server.printP(eventformstart);
    }

    P(eventtablestart) =
      "<TABLE border=1>"
      "<TR><TH colspan=3>Event Schedule</TH></TR>";
    server.printP(eventtablestart);

    uint8_t last_event = MAX_X10EVENTS - 1;
    for (i = MAX_X10EVENTS-1; i >= 0; --i) {
#if WEB_OUT_DEBUG || DEBUG_OUTP_EV
	DVFn("checking event #", i);
#endif
	if(config_data.x10events[i].hour != 0
	|| config_data.x10events[i].uidx != 0
	|| config_data.x10events[i].action != 2) {
	    last_event = (i < (MAX_X10EVENTS - 2)) ? ++i : MAX_X10EVENTS - 1 ;
#if WEB_OUT_DEBUG || DEBUG_OUTP_EV
	    DVFn("next after live event #", last_event);
#endif
	    break;
	}
    }
    for (i = 0; i <= last_event; ++i) {
#if WEB_OUT_DEBUG || DEBUG_OUTP_EV
	DVFn("start of event #", i);
#endif
	// hour uidx action
	// config_data.x10events[i].hour
	// config_data.x10events[i].uidx
	// config_data.x10events[i].action
	server.printP(PSTR("<TR><TD>"));
	if (addControls) {
	    char tName[4];
	    uint8_t j;
	    tName[0] = 't';
	    itoa(i, tName + 1, 10);
	    // select dropdown of 0..23,SR,SS
	    server.printP(PSTR("<select name=\""));
	    server << tName;
	    server.printP(PSTR("\">"));
	    for(j = 0 ; j < 24 ; j++) {
		server.printP(PSTR("<option value="));
		server.print(j);
		if(config_data.x10events[i].hour == j) server.printP(PSTR(" selected"));
		server.printP(PSTR(">"));
		server.print(j);
		server.printP(PSTR(":00</option>"));
	    }
	    for(j = 32 ; j < 34 ; j++) {
		server.printP(PSTR("<option value="));
		server.print(j);
		if(config_data.x10events[i].hour == j) server.printP(PSTR(" selected"));
		server.printP(PSTR(">"));
		if(j == 32) server.printP(PSTR("SUNRISE"));
		if(j == 33) server.printP(PSTR("SUNSET"));
		server.printP(PSTR("</option>"));
	    }
	    server.printP(PSTR("</select>"));
	}
	else {
	    if(config_data.x10events[i].hour < 32) {
		server.print(config_data.x10events[i].hour);
	    }
	    else {
		server.printP((config_data.x10events[i].hour == 32) ? PSTR("SR") : PSTR("SS"));
	    }
	}
	server.printP(PSTR("</TD>\n<TD>"));
	if (addControls) {
	    char uName[4];
	    int j;
	    uName[0] = 'u';
	    itoa(i, uName + 1, 10);
	    // select box of all unit names
	    server.printP(PSTR("<select name=\""));
	    server << uName;
	    server.printP(PSTR("\">"));
	    for(j = 0 ; j < MAX_HOUSE_CODES * 16 ; j++) {
		if(!config_data.x10devices[j].name[0]) { continue; }
		server.printP(PSTR("<option value=\""));
		server.print(j);
		server.printP(PSTR("\""));
		if(config_data.x10events[i].uidx == j) server.printP(PSTR(" selected"));
		server.printP(PSTR(">"));
		server.print(config_data.x10devices[j].name);
		server.printP(PSTR("</option>"));
	    }
	    server.printP(PSTR("</select>"));
	}
	else {
	    if(config_data.x10devices[config_data.x10events[i].uidx].name[0]) {
		server.print(config_data.x10devices[config_data.x10events[i].uidx].name);
	    }
	    server.printP(PSTR("&nbsp;"));
	}
	server.printP(PSTR("</TD>\n<TD>"));
	if (addControls) {
#if WEB_OUT_DEBUG || DEBUG_OUTP_EV
	    sprintf_P(Dbuf, PSTR("event %d, unit = %s, action = %d"), i, config_data.x10devices[config_data.x10events[i].uidx].name, config_data.x10events[i].action);
	    Serial.println(Dbuf);
#endif
	    // should be multiple choice of ON,OFF
	    char aName[4];
	    aName[0] = 'a';
	    itoa(i, aName + 1, 10);
	    server.printP(PSTR("<select name=\""));
	    server << aName;
	    server.printP(PSTR("\">"));
	    server.printP(PSTR("<option value=\""));
	    server.printP(PSTR("0\""));
	    if(!config_data.x10events[i].action) server.printP(PSTR(" selected"));
	    server.printP(PSTR(">OFF</option>"));
	    server.printP(PSTR("<option value=\""));
	    server.printP(PSTR("1\""));
	    if(config_data.x10events[i].action == 1) server.printP(PSTR(" selected"));
	    server.printP(PSTR(">ON</option>"));
	    server.printP(PSTR("<option value=\""));
	    server.printP(PSTR("2\""));
	    if(config_data.x10events[i].action == 2) server.printP(PSTR(" selected"));
	    server.printP(PSTR(">NOOP</option>"));
	    server.printP(PSTR("</select>"));
	}
	else {
	    server.printP(config_data.x10events[i].action == 2 ? PSTR("NOOP") : config_data.x10events[i].action == 1 ? PSTR("ON") : PSTR("OFF"));
	}
	server.printP(PSTR("</TD></TR>\n"));
#if WEB_OUT_DEBUG || DEBUG_OUTP_EV
	DVFn("end event #", i);
#endif
    }
    server.printP(PSTR("</table><p>"));

    if (addControls && system_state == IDLE) server.printP(formend);
    server.printP(root_link);

    server.printP(htmlTail);
#if WEB_OUT_DEBUG || DEBUG_OUTP_EV
    DLFn("end outputEvents()");
#endif
}

void outputHouseCodes(WebServer &server, WebServer::ConnectionType type, bool addControls = false) {

  int i;
  server.httpSuccess();
  server.printP(htmlHead);

  if (addControls) {
    server.printP(houseformstart);
  }

  P(housetablestart) =
      "<TABLE border=1>"
      "<TR><TH>House Codes</TH></TR>";
  server.printP(housetablestart);

  for (i = 0; i <= 1; i++) {
    server << "<TR><TD>";
    char fieldvalue[2];
    fieldvalue[0] = 'A' + config_data.house_codes[i];
    fieldvalue[1] = '\0';
    if (addControls) {
	char fieldname[4];
	fieldname[0] = 'h';
	itoa(i, fieldname + 1, 10);
	server.printP(PSTR("house "));
	server << i;
	server.printP(PSTR("<input type=\"text\" size=\"1\" maxlength=\"1\" name=\""));
	server << fieldname;
	server.printP(PSTR("\" value=\""));
	server << fieldvalue;
	server.printP(PSTR("\">"));
    }
    else {
	server << fieldvalue;
    }
    server.printP(PSTR("</TD></TR>"));
  }


  server.printP(PSTR("</table><p>"));

  if (addControls && system_state == IDLE) server.printP(formend);
  server.printP(root_link);

  server.printP(htmlTail);
#if WEB_OUT_DEBUG
    DLFn("end outputHouseCodes()");
#endif
}

#define DEBUG_OUT_UNIT_CODES 0
void outputUnitCodes(WebServer &server, WebServer::ConnectionType type, bool addControls = false) {

#if WEB_OUT_DEBUG || DEBUG_OUT_UNIT_CODES
  DLFn("begin outputUnitCodes()");
#endif
  server.httpSuccess();
  server.printP(htmlHead);

    if (addControls) {
	server.printP(unitformstart);
    }

    P(unittablestart) =
      "<TABLE border=1>"
      "<TR><TH colspan=3>House Codes</TH></TR>\n";
    server.printP(unittablestart);

    for (uint8_t h = 0; h < 2; h++) {
	char housevalue[2];
#if DEBUG_OUT_UNIT_CODES
	DVF("h =", h);
#endif
	housevalue[0] = 'A' + config_data.house_codes[h];
	housevalue[1] = '\0';
#if DEBUG_OUT_UNIT_CODES
	DVFn(", housevalue =", housevalue);
#endif
	for (uint8_t i = 0; i < 16; ++i) {
#if DEBUG_OUT_UNIT_CODES
	    DVF("i =", i);
#endif
	    // name housecode unitcode
	    server.printP(PSTR("<TR><TD>"));
	    if (addControls) {
		char fieldname[4];
		fieldname[0] = 'u';
		itoa((h*16)+i, fieldname + 1, 10);
#if DEBUG_OUT_UNIT_CODES
		DVF(", fieldname =", fieldname);
#endif
		server.printP(PSTR("<input type=\"text\" size=\"16\" maxlength=\"16\" name=\""));
		server << fieldname;
		server.printP(PSTR("\""));
		if(config_data.x10devices[(h * 16) + i].name[0]) {
		    server.printP(PSTR(" value=\""));
		    server << config_data.x10devices[(h * 16) + i].name;
		    server.printP(PSTR("\""));
#if DEBUG_OUT_UNIT_CODES
		    DVF(", lamp name =", config_data.x10devices[(h * 16) + i].name);
#endif
		}
		server.printP(PSTR("\">&nbsp;"));
#if DEBUG_OUT_UNIT_CODES
		Serial.println();
#endif
	    }
	    else {
#if DEBUG_OUT_UNIT_CODES
		Serial.println();
#endif
		if(config_data.x10devices[(h * 16) + i].name[0]) {
		    server << config_data.x10devices[(h * 16) + i].name;
		}
		else {
		    server.printP(PSTR("&nbsp;"));
		}
	    }
	    server << "</TD><TD>" << housevalue << "</TD><TD>" << i+1 << "</TD></TR>\n";
	}
    }
    server.printP(PSTR("</table><p>\n"));
    if (addControls && system_state == IDLE) server.printP(formend);
    server.printP(root_link);
    server.printP(htmlTail);
#if WEB_OUT_DEBUG || DEBUG_OUT_UNIT_CODES
    DLFn("end outputUnitCodes()");
#endif
}

void output_alarm(WebServer &server, WebServer::ConnectionType type, bool addControls = false) {
    show_alarm_page(server, type, addControls && (system_state == IDLE));
#if WEB_OUT_DEBUG
    DLFn("end output_alarm()");
#endif
}

void output_alarm_log(WebServer &server, WebServer::ConnectionType type, bool addControls = false) {
    show_almlog_page(server, type, addControls);
#if WEB_OUT_DEBUG
    DLFn("end output_alarm()");
#endif
}

void alarm_Cmd(WebServer &server, WebServer::ConnectionType type, char *url_tail, bool tail_complete) {
#define WEB_IN_DEBUG 0
#if WEB_IN_DEBUG
char sbuf[40];
#endif
    if (type == WebServer::POST) {
    bool repeat;
    char name[10], value[10];
    int zone;
    unsigned char sp;
	do {
	    repeat = server.readPOSTparam(name, 10, value, 10);
#define broken_parser 1
	    if(name[0] == 'b' && name[1] == 'l') {	// set backlight (bl0...) _BL
		zone = name[2] - '0';
		int val = strtoul(value, NULL, 10);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("st[%d] = '%s'"), zone, value);
		Serial.println(Dbuf);
		sprintf_P(Dbuf, PSTR("old = %d, new = %d"), (int)config_data.backlight, (int)val);
		Serial.println(Dbuf);
#endif
		if(config_data.backlight == val) {	// no change
		    continue;
		}
		config_data.backlight = val;
		LCDbacklight(zone, (unsigned int)config_data.backlight);
		// display[i].backlight_timer = BLTIMEOUT;
	    }
	    else if(name[0] == 'b' && name[1] == 'd') {	// set dimbacklight (bd0...) _BD
		zone = name[2] - '0';
		int val = strtoul(value, NULL, 10);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("st[%d] = '%s'"), zone, value);
		Serial.println(Dbuf);
		sprintf_P(Dbuf, PSTR("old = %d, new = %d"), (int)config_data.dimbacklight, (int)val);
		Serial.println(Dbuf);
#endif
		if(config_data.dimbacklight == val) {	// no change
		    continue;
		}
		config_data.dimbacklight = val;
		LCDbacklight(zone, (unsigned int)config_data.dimbacklight);
	    }
	    else if(name[0] == 'a' && name[1] == 'u') {	// "auma0" (?AUMA)
		zone = name[4] - '0';
		char val = value[0];	// legal values 'e' and 'd'
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("auma[%d] = '%s' = '%c', was '%c'"), zone, value, val, alarm_config.zone_config[zone].enabled);
		Serial.println(Dbuf);
#endif
		if(system_state == IDLE
		&& alarm_config.zone_config[zone].enabled != val) {
		    alarm_config.zone_config[zone].enabled = val;
		}
	    }
	    else if(name[0] == 's' && name[1] == 'c') {	// save config (scfg.) (?SAVE)
		zone = name[4] - '0';
		int val = strtoul(value, NULL, 10);
#if WEB_IN_DEBUG
		DVF("scfg: name", name);
		DVFn(", zone", zone);
		sprintf_P(Dbuf, PSTR("scfg[%d] = '%s' = %d"), zone, value, val);
		Serial.println(Dbuf);
#endif
		if(val == 1) {
		    eeprom_save_all();
		}
	    }
	    else if(!strcmp("ARM-PASSWD", name)) {	// arm system
		if(ui_mode[0] == UI_IDLE && !open_enabled_vector) {
		    play_sound(TADAAA);
		    set_system_wait();
		}
	    }
	    else if(!strcmp("DISARM-PWD", name)) {	// disarm system
		if(ui_mode[0] == UI_ARMED) {
		    disarm_system();
		    // play other happy sound
		    play_sound(NOTES);
		}
	    }
#if !broken_parser
	    else if (strstr_P(Dbuf2,PSTR("GET /reset")) != 0){ // remote reset, just for the heck of it
		resetFunc();
	    }
#endif
	} while (repeat);
	server.httpSeeOther(PREFIX "/alarm");
    }
    else {
	output_alarm(server, type, true);
    }
}

void almlog_Cmd(WebServer &server, WebServer::ConnectionType type, char *url_tail, bool tail_complete) {
    if (type != WebServer::POST) {
	output_alarm_log(server, type, false);
    }
}

// alter configuration settings from web form
void configCmd(WebServer &server, WebServer::ConnectionType type, char *url_tail, bool tail_complete) {
#if WEB_IN_DEBUG
char sbuf[30];
#endif
    if (type == WebServer::POST) {
    bool repeat, change=false;
    char name[11], value[30];
	do {
	    repeat = server.readPOSTparam(name, 11, value, 30);
	    if (name[0] == 'T' && name[1] == 'Z') {	// timezone
		int val = strtoul(value, NULL, 10);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("Z = '%s'"), value);
		Serial.println(Dbuf);
		sprintf_P(Dbuf, PSTR("old = %d, new = %d"), (int)config_data.tzindex, (int)val);
		Serial.println(Dbuf);
#endif
		if(config_data.tzindex == (TZindex)val) {	// no change
		    continue;
		}
		config_data.tzindex = (TZindex)val;
		change = true;
	    }
	    else if(!strcmp("battvolts", name)) { // battery A/D calibration
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		// constant = (batt10mvolts << 10) / average_batt_reading
		int battv = strtoul(value, NULL, 10);
		battv += 5;  // so we will round correctly when we divide by 10
		battv /= 10;	// get rid of trailing 0
		unsigned long batt_cal = ((long)battv << 10) / (long)average_batt_reading;
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("V = %d x 10mV"), battv);
		Serial.println(Dbuf);
#endif
		if(batt_cal > 1480 && batt_cal < 1580
		&& config_data.batt_cal_constant != batt_cal) {
		    config_data.batt_cal_constant = batt_cal;
		    change = true;
		}
	    }
	    else if(name[0] == 'm'
		 && name[1] >= '0'
		 && name[1] <= '5') {	// mac address m0,m1,m2,m3,m4,m5
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		int mindex = strtoul(name + 1, NULL, 10);
		int mvalue = strtoul(value, NULL, 16);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("field %d = %d"), mindex, mvalue);
		Serial.println(Dbuf);
#endif
		if(mvalue < 256) {
		    if(config_data.mac[mindex] != mvalue) {
			config_data.mac[mindex] = mvalue;
			change = true;
		    }
		}
	    }
	    else if(name[0] == 'i'
		 && name[1] >= '0'
		 && name[1] <= '3') {	// ip address i0,i1,i2,i3
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		int iindex = strtoul(name + 1, NULL, 10);
		int ivalue = strtoul(value, NULL, 10);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("field %d = %d"), iindex, ivalue);
		Serial.println(Dbuf);
#endif
		if(ivalue < 256) {
		    if(config_data.ip[iindex] != ivalue) {
			config_data.ip[iindex] = ivalue;
			change = true;
		    }
		}
	    }
	    else if(name[0] == 'd'
		 && name[1] >= '0'
		 && name[1] <= '3') {	// dns address d0,d1,d2,d3
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		int iindex = strtoul(name + 1, NULL, 10);
		int ivalue = strtoul(value, NULL, 10);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("field %d = %d"), iindex, ivalue);
		Serial.println(Dbuf);
#endif
		if(ivalue < 256) {
		    if(config_data.dns_server[iindex] != ivalue) {
			config_data.dns_server[iindex] = ivalue;
			change = true;
		    }
		}
	    }
	    else if(name[0] == 'g'
		 && name[1] >= '0'
		 && name[1] <= '3') {	// gateway address g0,g1,g2,g3
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		int iindex = strtoul(name + 1, NULL, 10);
		int ivalue = strtoul(value, NULL, 10);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("field %d = %d"), iindex, ivalue);
		Serial.println(Dbuf);
#endif
		if(ivalue < 256) {
		    if(config_data.gateway_ip[iindex] != ivalue) {
			config_data.gateway_ip[iindex] = ivalue;
			change = true;
		    }
		}
	    }
	    else if(name[0] == 's'
		 && name[1] >= '0'
		 && name[1] <= '3') {	// subnet mask s0,s1,s2,s3
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		int iindex = strtoul(name + 1, NULL, 10);
		int ivalue = strtoul(value, NULL, 10);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("field %d = %d"), iindex, ivalue);
		Serial.println(Dbuf);
#endif
		if(ivalue < 256) {
		    if(config_data.subnet_mask[iindex] != ivalue) {
			config_data.subnet_mask[iindex] = ivalue;
			change = true;
		    }
		}
	    }
	    else if(!strcmp("ssid", name)) { // wifi ssid
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		// if value not "", not "****", and not same as ssid
		if(strlen(value) && strncmp(value, "****", 4) && strcmp(config_data.wifi_ssid, value)) {
		    strncpy(config_data.wifi_ssid, value, sizeof(config_data.wifi_ssid));
		    change = true;
		}
	    }
	    else if(!strcmp("pass", name)) { // wifi password
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		// if value not "", not "****", and not same as pass
		if(strlen(value) && strncmp(value, "****", 4) && strcmp(config_data.wifi_pass, value)) {
		    strncpy(config_data.wifi_pass, value, sizeof(config_data.wifi_ssid));
		    change = true;
		}
	    }
	    else if(!strcmp("nmen", name)) { // night mode enabled
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		int ivalue = strtoul(value, NULL, 10);
		if((ivalue == 0 || ivalue == 1)
		&& alarm_config.night_mode_enabled != ivalue) {
		    alarm_config.night_mode_enabled = ivalue;
		    change = true;
		}
	    }
	    else if(!strcmp("nmsh", name)) { // night mode start hour
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		int ivalue = strtoul(value, NULL, 10);
		if(ivalue >= 0 && ivalue < 24
		&& alarm_config.night_start[1] != ivalue) {
		    alarm_config.night_start[1] = ivalue;
		    change = true;
		}
	    }
	    else if(!strcmp("nmsm", name)) { // night mode start minute
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		int ivalue = strtoul(value, NULL, 10);
		if(ivalue >= 0 && ivalue < 60
		&& alarm_config.night_start[0] != ivalue) {
		    alarm_config.night_start[0] = ivalue;
		    change = true;
		}
	    }
	    else if(!strcmp("nmeh", name)) { // night mode stop hour
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		int ivalue = strtoul(value, NULL, 10);
		if(ivalue >= 0 && ivalue < 24
		&& alarm_config.night_stop[1] != ivalue) {
		    alarm_config.night_stop[1] = ivalue;
		    change = true;
		}
	    }
	    else if(!strcmp("nmem", name)) { // night mode stop minute
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		int ivalue = strtoul(value, NULL, 10);
		if(ivalue >= 0 && ivalue < 60
		&& alarm_config.night_stop[0] != ivalue) {
		    alarm_config.night_stop[0] = ivalue;
		    change = true;
		}
	    }
	} while (repeat);
	server.httpSeeOther(PREFIX "/config");
	if(change) {
	    EEPROM_writeAnything(sizeof(CONFIG_ALARM), config_data);
	}
    }
    else {
	outputConfig(server, type, true);
    }
}

void manualCmd(WebServer &server, WebServer::ConnectionType type, char *url_tail, bool tail_complete) {
#if WEB_IN_DEBUG
char sbuf[50];
#endif
    if (type == WebServer::POST) {
    bool repeat;
    char name[5], value[5];
	do {
	    repeat = server.readPOSTparam(name, 5, value, 5);
	    if (name[0] == 'L') {	// for example, name= L13 , value= 2
		int index = strtoul(name + 1, NULL, 10);
		int val = strtoul(value, NULL, 10);
		if(val == 2) {	// no-op
		    continue;
		}
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("L%d = '%s'"), index, value);
		Serial.println(Dbuf);
		sprintf_P(Dbuf, PSTR("i = %d, name = '%s', command = '%s'"), index, config_data.x10devices[index].name, value);
		Serial.println(Dbuf);
#endif
		// uint8_t evhouse  = config_data.house_codes[index >> 4];
		// uint8_t evunit   = index & 0x0f;
		action_queue[action_ctr].evidx = index;
		action_queue[action_ctr].action = val;
		action_queue[action_ctr++].ctr = 4;
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("adding %d %d %d to action_queue[%d]"), index, val, 4, action_ctr-1);
		Serial.println(Dbuf);
#endif
	    }
	} while (repeat);
	server.httpSeeOther(PREFIX "/manual");
    }
    else {
	outputManual(server, type, true);
    }
}

void houseCmd(WebServer &server, WebServer::ConnectionType type, char *url_tail, bool tail_complete) {
URLPARAM_RESULT rc;
#if WEB_IN_DEBUG
char sbuf[30];
#endif

    if (type == WebServer::POST) {
	bool repeat;
	char name[5], value[5];
	do {
	    repeat = server.readPOSTparam(name, 5, value, 5);
	    if (name[0] == 'h') {
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = %s"), name, value);
		Serial.println(Dbuf);
#endif
		int hindex = strtoul(name + 1, NULL, 10);
		int hvalue = value[0] - 'A';
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%d = %d"), hindex, hvalue);
		Serial.println(Dbuf);
#endif
		if(hvalue < 16) {
		    if(config_data.house_codes[hindex] != hvalue) {
			config_data.house_codes[hindex] = hvalue;
			EEPROM_writeAnything(sizeof(CONFIG_ALARM), config_data);
		    }
		}
	    }
	} while (repeat);
	server.httpSeeOther(PREFIX "/housecodes");
    }
    else {
	outputHouseCodes(server, type, true);
    }
}

void unitCmd(WebServer &server, WebServer::ConnectionType type, char *url_tail, bool tail_complete) {
#if WEB_IN_DEBUG
char sbuf[50];
#endif
    if (type == WebServer::POST) {
	bool repeat;
	char name[5], value[20];
	bool change=false;
	do {
	    repeat = server.readPOSTparam(name, 5, value, 17);
	    if (name[0] == 'u') {
		int index = strtoul(name + 1, NULL, 10);
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("u%d = '%s'"), index, value);
		Serial.println(Dbuf);
#endif
		if(strncmp(config_data.x10devices[index].name, value, 16)) {
#if WEB_IN_DEBUG
		    sprintf_P(Dbuf, PSTR("different: i = %d, old = '%s', new = '%s'"), index, config_data.x10devices[index].name, value);
		    Serial.println(Dbuf);
#endif
		    strncpy(config_data.x10devices[index].name, value, 16);
		    config_data.x10devices[index].name[16] = '\0';
		    change = true;
		}
	    }
	} while (repeat);
	server.httpSeeOther(PREFIX "/unit_codes");
	if(change) {
	    EEPROM_writeAnything(sizeof(CONFIG_ALARM), config_data);
	}
    }
    else {
	outputUnitCodes(server, type, true);
    }
}

void schedCmd(WebServer &server, WebServer::ConnectionType type, char *url_tail, bool tail_complete) {
#if WEB_IN_DEBUG
char sbuf[50];
#endif
    if (type == WebServer::POST) {
	bool repeat, change;
	char name[5], value[16];
	change = 0;
	do {
	    repeat = server.readPOSTparam(name, 5, value, 16);
	    int idx = strtoul(name + 1, NULL, 10);
	    int val = strtoul(value, NULL, 10);
	    if (name[0] == 't') {	/* time */
		if(config_data.x10events[idx].hour != val) {
		    config_data.x10events[idx].hour = val;
		    change = 1;
		}
	    }
	    if (name[0] == 'u') {	/* unit */
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = '%s'"), name, value);
		Serial.println(Dbuf);
#endif
		if(config_data.x10events[idx].uidx != val) {
#if WEB_IN_DEBUG
		    sprintf_P(Dbuf, PSTR("i = %d, old = %d, new = %d"), idx, config_data.x10events[idx].uidx, val);
		    Serial.println(Dbuf);
#endif
		    config_data.x10events[idx].uidx = val;
		    change = 1;
		}
	    }
	    if (name[0] == 'a') {	/* action */
#if WEB_IN_DEBUG
		sprintf_P(Dbuf, PSTR("%s = '%s'"), name, value);
		Serial.println(Dbuf);
#endif
		if(config_data.x10events[idx].action != val) {
#if WEB_IN_DEBUG
		    sprintf_P(Dbuf, PSTR("i = %d, old = %d, new = %d"), idx, config_data.x10events[idx].action, val);
		    Serial.println(Dbuf);
#endif
		    config_data.x10events[idx].action = val;
		    change = 1;
		}
	    }
	} while (repeat);
	if(change) {
	    EEPROM_writeAnything(sizeof(CONFIG_ALARM), config_data);
	}
	server.httpSeeOther(PREFIX "/schedule");
    }
    else {
	outputEvents(server, type, true);
    }
}

void defaultCmd(WebServer &server, WebServer::ConnectionType type, char *url_tail, bool tail_complete) {
int i;
char sname[20];
  server.httpSuccess();
  server.printP(htmlHead);

  server << "<h1>Alarm</h1><p>";
  // show system state
  strncpy_P(sname, (char*)pgm_read_word(&(sstate_names[(int)system_state])), sizeof(sname));
  server << "<h4>";
  server << sname;
  server << "</h4>";

  // show battery voltage
  server << "battery: ";
  unsigned int v = ((long)average_batt_reading * (long)config_data.batt_cal_constant) >> 10;
  DVFn("raw v =", analogRead(A8));
  DVFn("ave v =", average_batt_reading);
  sprintf_P(sname, PSTR("%d.%02dV<br>\n"), v/100, v%100);
  server.print(sname);

  // show uptime
  unsigned int updays = uptime_min/1440L ;
  unsigned int uphrs = (uptime_min % 1440L)/60 ;
  unsigned int upmins = uptime_min % 60L ;
  server.printP(uptimestart);
  sprintf_P(sname, PSTR("%4d days %02d:%02d<br><br>"), updays, uphrs, upmins);
  server.print(sname);

  server.printP(alarm_link);
  server.printP(almlog_link);
  server << "</p><h1>X10 Scheduler</h1><p>";
  server.printP(houselink);
  server.printP(unitslink);
  server.printP(schedlink);
  server.printP(manuallink);
  server << "</p><h1>General</h1><p>";
  server.printP(configlink);

  server.printP(angle_br);
  server.printP(htmlTail);
}

void send_command(char house, unsigned char unit, unsigned char action) {
uint8_t debug=0;
    if(debug) {
	sprintf_P(Dbuf, PSTR("cm17control(%c, %2d, %s)"), (char)('A' + house), 1 + unit, action ? "on" : "off");
	Serial.println(Dbuf);
    }
    X10.sendCmd( (HouseCode)house, (int)(unit+1), (CommandCode)!action);
}

// http://playground.arduino.cc/Code/Email

byte eRcv (void) {
byte respCode;
byte thisByte;
int loopCount = 0;
byte eRmode = 0;

PassTwo:
    while(!client.available()) {
	delay(1);	// 1 millisecond
	my_yield();	// poll the user interface
	loopCount++;

	// if nothing received for 10 seconds, timeout
	if(loopCount > 10000) {
	    client.stop();
	    Serial.println(F("\r\nTimeout"));
	    return 0;
	}
    }
    // DVFn("eRcv loopCount =", loopCount);

    if(!eRmode) {
	respCode = client.peek();
	DHFn("respCode =", respCode);
    }

    // DLFn("before read everything");
    while(client.available()) {  
	thisByte = client.read();    
	Serial.write(thisByte);
    }
    // DLFn("after read everything");

    if(eRmode) {
	client.stop();
	Serial.println(F("disconnected"));
	return(0);
    }
    else if(respCode >= '4') {
	eRmode = 1;
	client.println(F("QUIT"));
	goto PassTwo;
    }
    return 1;
}

#define TIME_SEND_EMAIL 0

byte sendEmail (uint8_t zone) {
byte thisByte = 0;
#if TIME_SEND_EMAIL
unsigned long et;

    et = millis();
#endif
char lzname[LONGEST_NAME];	// max length of any string in zone_names[]
    if(client.connect(mailserver,25) == 1) {
	Serial.println(F("connected"));
    } else {
	Serial.println(F("connection failed"));
	return 0;
    }
    if(!eRcv()) return 0;
#if TIME_SEND_EMAIL
    DVFn("after connect:", millis() - et);
#endif

    Serial.print(F("HELO" ));
    Serial.println(mailclientname);
    client.print(F("HELO "));
    client.println(mailclientname);
    if(!eRcv()) return 0;
#if TIME_SEND_EMAIL
    DVFn("after HELO:", millis() - et);
#endif

    Serial.println(F("Sending From"));
    // change to your email address (sender)
    client.println(F("MAIL From: <bill@dudley.nu>"));
    if(!eRcv()) return 0;
#if TIME_SEND_EMAIL
    DVFn("after From:", millis() - et);
#endif

    // change to recipient address
    DVF("RCPT To: <", recipient_name);
    DLFn(">");
    client.print(F("RCPT To: <"));
    client.print(recipient_name);
    client.println(F(">"));
    if(!eRcv()) return 0;
#if TIME_SEND_EMAIL
    DVFn("after RCPT To:", millis() - et);
#endif

    Serial.println(F("Sending DATA"));
    client.println(F("DATA"));
    if(!eRcv()) return 0;
#if TIME_SEND_EMAIL
    DVFn("after DATA:", millis() - et);
#endif

    Serial.println(F("Sending email"));
    // change to recipient address
    client.print(F("To: <"));
    client.print(recipient_name);
    client.println(F(">"));
#if TIME_SEND_EMAIL
    DVFn("after To:", millis() - et);
#endif

    client.print(F("From: "));
    client.println(from_email_addr);
    client.println(F("Subject: Alarm System Alert\r\n"));
    client.print(F("This is from the Alarm System at "));
    client.print(alarm_location);
    client.println(F("."));
    client.print(F("Violation in zone: "));
    strncpy_P(lzname, (char*)pgm_read_word(&(zone_names[zone])), sizeof(lzname));
    client.println(lzname);
    client.println(F("."));
#if TIME_SEND_EMAIL
    DVFn("after Body:", millis() - et);
#endif

    if(!eRcv()) return 0;

    Serial.println(F("Sending QUIT"));
    client.println(F("QUIT"));
#if TIME_SEND_EMAIL
    DVFn("after Quit:", millis() - et);
#endif

    if(!eRcv()) return 0;

    client.stop();

    Serial.println(F("disconnected"));
#if TIME_SEND_EMAIL
    DVFn("after client.stop:", millis() - et);
#endif

    return 1;
}

byte eRcvWiFi (void) {
byte gotChars = 0;
char rxbuf[256];
int loopCount = 0;
int rxlen = 0;

    for(loopCount = 0 ; loopCount < 10 ; loopCount++) {
	memset(rxbuf, '\0', sizeof(rxbuf));
	rxlen = wifi.recv((uint8_t *)rxbuf, sizeof(rxbuf), 1000); // wait 1 sec max
	my_yield();	// poll the user interface, honk watchdog timer
	if(rxlen > 0) {	// length of reply is non-zero
	    gotChars++;
	    Serial.print(rxbuf);	// echo to debug (optional)
	}
	else {	// timed out
	    if(gotChars) {	// but we already have a reply
		return 1;	// so this was a success
	    }
	}
    }
    if(!gotChars) {		// then 10 timeouts occurred with no reply
	Serial.println(F("\nTimeout"));
	wifi.println(F("QUIT"));
	wifi.releaseTCP();
	Serial.println(F("disconnected"));
	return(0);
    }
    return 1;
}

byte sendEmailWiFi (uint8_t zone) {
byte thisByte = 0;
byte buffer[256];
char lzname[LONGEST_NAME];	// max length of any string in zone_names[]
#if TIME_SEND_EMAIL
unsigned long et;

    et = millis();
#endif
    if(wifi.createTCP(mailserverWiFi,25) == 1) {
	Serial.println(F("connected"));
    } else {
	Serial.println(F("connection failed"));
	return 0;
    }
    if(!eRcvWiFi()) return 0;	// wait for 220 message
#if TIME_SEND_EMAIL
    DVFn("after connect:", millis() - et);
#endif
    Serial.print(F("HELO "));
    Serial.println(mailclientname);
    wifi.print(F("HELO"));
    wifi.println(mailclientname);
    if(!eRcvWiFi()) return 0;	// wait for 250 message
#if TIME_SEND_EMAIL
    DVFn("after HELO:", millis() - et);
#endif
    Serial.println(F("Sending From"));
    // change to your email address (sender)
    wifi.println(F("MAIL From: <bill@dudley.nu>"));
    if(!eRcvWiFi()) return 0;	// wait for 250 message
#if TIME_SEND_EMAIL
    DVFn("after From:", millis() - et);
#endif
    // change to recipient address
    DVF("RCPT To: <", recipient_name);
    DLFn(">");
    // wifi.print(F("RCPT To: <"));
    wifi.print(F("RCPT To: <"));
    wifi.print(recipient_name);
    wifi.println(F(">"));
    if(!eRcvWiFi()) return 0;
#if TIME_SEND_EMAIL
    DVFn("after RCPT To:", millis() - et);
#endif

    Serial.println(F("Sending DATA"));
    wifi.println(F("DATA"));
    if(!eRcvWiFi()) return 0;	// wait for: 354 Enter mail, end with "."
#if TIME_SEND_EMAIL
    DVFn("after DATA:", millis() - et);
#endif

    Serial.println(F("Sending email"));
    // change to recipient address
    wifi.print(F("To: <"));
    wifi.print(recipient_name);
    wifi.println(F(">"));
#if TIME_SEND_EMAIL
    DVFn("after To:", millis() - et);
#endif

    wifi.print(F("From: "));
    wifi.println(from_email_addr);
    wifi.println(F("Subject: Alarm System Alert\r\n"));
    wifi.print(F("This is from the Alarm System at "));
    wifi.print(alarm_location);
    wifi.println(".");
    wifi.print(F("Violation in zone: "));
    strncpy_P(lzname, (char*)pgm_read_word(&(zone_names[zone])), sizeof(lzname));
    wifi.println(lzname);
    wifi.println(".");
#if TIME_SEND_EMAIL
    DVFn("after Body:", millis() - et);
#endif

    if(!eRcvWiFi()) return 0;	// wait for 250 message accepted for delivery

    Serial.println(F("Sending QUIT"));
    wifi.println(F("QUIT"));
#if TIME_SEND_EMAIL
    DVFn("after Quit:", millis() - et);
#endif

    if(!eRcvWiFi()) return 0;	// wait for 221 . . . closing connection

    wifi.releaseTCP();

    Serial.println(F("disconnected"));
#if TIME_SEND_EMAIL
    DVFn("after wifi.releaseTCP:", millis() - et);
#endif

    return 1;
}

void alarm_bell_on (void) {
    DLFn("Alarm Relay ON");
    digitalWrite(ALARM_RELAY, LOW);
}

void alarm_bell_off (void) {
    DLFn("Alarm Relay OFF");
    digitalWrite(ALARM_RELAY, HIGH);
}

void wifi_setup (void) {
uint8_t esp8266_ok;
    DLF("ESP8266 FW Version: ");
    Serial.println(wifi.getVersion().c_str());

    if (wifi.setOprToStation()) {
        DLFn("to station ok");
	esp8266_ok = 1;
    } else {
        DLFn("to station err");
	esp8266_ok = 0;
    }

    wifi_connected = 0;
    if(esp8266_ok) {	// no point in this if esp8266 is unhappy
	for(byte wc = 0 ; wc < MAX_WIFI_TRIES ; wc++) {
	    if (wifi.joinAP(config_data.wifi_ssid, config_data.wifi_pass)) {
		DLFn("Join AP success");
		DLF("IP: ");       
		Serial.println(wifi.getLocalIP().c_str());
		wifi_connected = 1;
		break;
	    } else {
		DLFn("Join AP failure");
	    }
	}
    }
    
    DLF("wifi ");
    if(wifi.disableMUX()) {
	DLFn("ok");
    }
    else {
	DLFn("err");
    }
}


time_t getNTPtimeAny (void) {
time_t r;
    DLFn("trying getNTPtime()");
    r = getNTPtime();
    if(r) return r;
    DLFn("trying getNTPtimeWifi()");
    r = getNTPtimeWiFi();
    if(r) return r;
    // WiFi NTP failed, so try to restart WiFi connection
    wifi_setup();
    r = getNTPtimeWiFi();
    if(r) return r;
    ntp_fail_counter++;	// internet is hosed
    DVFn("config_data.clock_is_set =", config_data.clock_is_set);
    if(config_data.clock_is_set) {
	DLFn("trying rtc.get()");
	r = rtc.get();
    }
    return r;
}

// my wrapper; try different NTP servers until we get a reply
time_t getNTPtimeWiFi (void) {
char ntp_host[35];
time_t tUTC;
    do {
	strncpy_P(ntp_host,
	    (char *)pgm_read_word(&(ntp_hosts[ntp_host_idx_wifi++])), 34);
	lcd.clear();
	lcd.print(ntp_host);
	tUTC = get_NTP_timeWiFi(ntp_host);
	// DVFn("get_NTP_timeWiFi =", tUTC);
	// rotate through all NTP hosts
	if(ntp_host_idx_wifi >= (sizeof(ntp_hosts)/sizeof(const char *))) {
	    ntp_host_idx_wifi = 0;
	}
	if(tUTC) {
	    lcd.clear();
	    // DVFn("getNTPtimeWiFi", __LINE__);
#if MIN_TIMERS_NEEDED
	    // schedule time check for almost a day from now
	    min_timers[NTP_MIN_TIMER] = 24 * 59;
#endif
	    DVFn("getNTPtimeWiFi tUTC =", tUTC);
	    return(tUTC);
	}
    } while(1);
    return (time_t)0;
}

/*
 Udp NTP Client
 
 Get the time from a Network Time Protocol (NTP) time server
 Demonstrates use of UDP sendPacket and ReceivePacket 
 For more on NTP time servers and the messages needed to communicate with them, 
 see http://en.wikipedia.org/wiki/Network_Time_Protocol
 
 created 4 Sep 2010 by Michael Margolis
 modified 9 Apr 2012 by Tom Igoe
 modified 10 Feb 2015 by Alan Shiflett
 modified 13 Jan 2016 for esp8266 by Bill Dudley
 This code is in the public domain.
*/

// get the UTC time from an NTP server via WiFi
time_t get_NTP_timeWiFi (const char * ntp_server) {
byte packetBuffer [NTP_PACKET_SIZE]; // for both incoming & outgoing packets 
uint32_t len;
time_t result;
    len = sendNTPpacketWiFi(ntp_server, packetBuffer); // send NTP packet to time server
    result = 0;
    if (len) {
	//the timestamp is at bytes 40-43 of the received packet
	// or two words, long. First, esxtract the two words:

	unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
	unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);  
	// combine the four bytes (two words) into a long integer
	// this is NTP time (seconds since Jan 1 1900):
	unsigned long secsSince1900 = highWord << 16 | lowWord;  

	// now convert NTP time into UNIX time:
	// Unix time starts on Jan 1 1970. In seconds, that's 2208988800:
	const unsigned long seventyYears = 2208988800UL;     
	// subtract seventy years:
	result = secsSince1900 - seventyYears;  
    }
    // DVFn("UTC Unix time ", result);
    return result;
}

// send an NTP request to the time server at the given address 
uint32_t sendNTPpacketWiFi (const char * address, byte *pbuffer) {
bool sok;
uint32_t len;
    DVFn("connecting to", address);
    sok = wifi.registerUDP(address, 123); //NTP requests are to port 123
    if(!sok) {
	Serial.println(F("registerUDP() err"));
    }

    // set all bytes in the buffer to 0
    memset(pbuffer, 0, NTP_PACKET_SIZE); 
    // Initialize values needed to form NTP request
    // (see URL above for details on the packets)
    pbuffer[0] = 0b11100011;   // LI, Version, Mode
    pbuffer[1] = 0;     // Stratum, or type of clock
    pbuffer[2] = 6;     // Polling Interval
    pbuffer[3] = 0xEC;  // Peer Clock Precision
    // 8 bytes of zero for Root Delay & Root Dispersion
    pbuffer[12]  = 49; 
    pbuffer[13]  = 0x4E;
    pbuffer[14]  = 49;
    pbuffer[15]  = 52;

    // all NTP fields have been given values, now
    // send a packet requesting a timestamp: 		   
    sok = wifi.send(pbuffer, NTP_PACKET_SIZE);
    if(sok) { DLFn("packet sent"); }

    // receive with 10 second timeout:
    for(uint8_t j = 0 ; j < 10 ; j++) {
	len = wifi.recv(pbuffer, NTP_PACKET_SIZE, 1000);
	my_yield();	// poll the user interface, honk watchdog timer
	if(len) break;
    }
    DVFn("received =", len);

    sok = wifi.unregisterUDP();
    if(!sok) { Serial.println(F("unregUDP() ERR")); }
    return len;
}
