#include "ds3231.h"

#define DS3231_I2C_ADDRESS 0x68

// empty constructor
DS3231::DS3231() { }

// empty destructor
DS3231::~DS3231() { }

time_t DS3231::get() {
tmElements_t tm;
time_t ts;
uint8_t goodread;
uint8_t tries = 0;
    do {
	goodread = 1;
	Wire.beginTransmission(DS3231_I2C_ADDRESS);
	Wire.write(0); // set DS3231 register pointer to 00h
	Wire.endTransmission();
	Wire.requestFrom(DS3231_I2C_ADDRESS, 7);
	// request seven uint8_ts of data from DS3231 starting from register 00h
	tm.Second = bcdToDec(Wire.read() & 0x7f);
	tm.Minute = bcdToDec(Wire.read());
	tm.Hour   = bcdToDec(Wire.read() & 0x3f);
	tm.Wday   = bcdToDec(Wire.read());
	tm.Day    = bcdToDec(Wire.read());
	tm.Month  = bcdToDec(Wire.read());
	tm.Year   = bcdToDec(Wire.read());
	ts = makeTime(tm);
	if(ts < 1455000000L) {
	    goodread = 0;
	    delay(2);
	}
    } while(!goodread && ++tries < 10);
    return(ts);
}

uint8_t DS3231::set(time_t t) {
tmElements_t tm;
    breakTime(t, tm);
    Wire.beginTransmission(DS3231_I2C_ADDRESS);
    Wire.write(0); // set next input to start at the seconds register
    Wire.write(decToBcd(tm.Second));             // Disable Clock halt flag
    Wire.write(decToBcd(tm.Minute));
    Wire.write(decToBcd(tm.Hour));               // 24-hour mode
    Wire.write(tm.Wday);
    Wire.write(decToBcd(tm.Day));
    Wire.write(decToBcd(tm.Month));
    Wire.write(decToBcd(tm.Year));
    Wire.endTransmission();
    return 0;	// for semi-compatibility with DS1302RTC library
}

// private functions
// Convert normal decimal numbers to binary coded decimal
uint8_t DS3231::decToBcd(uint8_t val) {
  return( (val/10*16) + (val%10) );
}

// Convert binary coded decimal to normal decimal numbers
uint8_t DS3231::bcdToDec(uint8_t val) {
  return( (val/16*10) + (val%16) );
}
