#include "alarm.h"
#define LCD_MAIN 1
#include "lcdisplay.h"
#include "LiquidCrystal_I2C.h"

extern LiquidCrystal_I2C lcd;

//
//  LCD display routines
//
/*  ?a    home cursor
    ?b    destructive backspace
    ?B    backlight intensity
    ?c0   turn off cursor
    ?D0................ define char 0 as bitmap
    ?g    beep
    ?G216 set 2x16 display Geometry
    ?G420 set 4x20 display Geometry
    ?h    non-destructive backspace
    ?i    move cursor forward
    ?I    send next char to LCD if in range 0 to 3
    ?j    line feed
    ?k    reverse line feed
    ?l    clear current line
    ?m    carriage return
    ?n    new line (CR, LF, clear)
    ?s    set horizontal tab (value 1 to 8)
    ?S.   set config.start_up to value in range 0 to 2 inclusive
          1 means show stock start screen on power up.
          2 means show custom start screen on power up.
    ?t    move forward one tab
    ?f    clear
    ?y.   row
    ?x..  column
    ?!.   send next char to lcd as command byte
    ?>.   turn on big characters, arg either 3 or 4
    ?<    turn off big characters
    ?R    restore default custom chars
    ?*    display custom screen
    ?C.   define custom screen - send line number followed by chars to display
    ------------------QUAD POT STUFF---------------------
    ?H.   set quad pot maximum (0 to 9 inclusive)
    ?L.   set quad pot minimum (0 to 9 inclusive)
    ADD THIS:
    ?Q.   set quad pot current value (0 to 9 inclusive)
 */

void LCDbacklight (unsigned char d, unsigned int intensity) {
HardwareSerial *hsp;
char buf[4];
    hsp = LCDportarray[d];
    (*hsp).print("?B");   // backlight intensity
    sprintf(buf, "%02X", intensity);
    (*hsp).print(buf);
    if(d == 1) {
	analogWrite(SECOND_BKLT, intensity);
    }
}

// define a "degree" symbol as character "?0"
void defineDEGsym (unsigned char d) {
HardwareSerial *hsp;
    hsp = LCDportarray[d];
    (*hsp).print("?D0040A040000000000");   // special character 0
}

void setupLCD (void) {
HardwareSerial *hsp;
unsigned char d;
    for(d = 0 ; d < MAXDISPLAYS ; d++) {
	hsp = LCDportarray[d];
	(*hsp).begin(9600);
	LCDbacklight(d, (unsigned int)config_data.backlight);
//	defineDEGsym(d);
	delay(100);
	(*hsp).print("?c0");    // No Cursor
	(*hsp).print("?G");	// sends "?G216" for 2x16 display
	(*hsp).print((int)(config_data.display[d].rows));
	(*hsp).print((int)(config_data.display[d].cols));
	(*hsp).print("?f");     //clear command.
	delay(100);
    }
}

void LCDline (unsigned char d, unsigned char line) {
HardwareSerial *hsp;
    hsp = LCDportarray[d];
    (*hsp).print("?y");   //Row #
    (*hsp).print((int)line);
    (*hsp).print("?x00");    //Column 0
    if(d == 1) {
	lcd.setCursor(0, line);
    }
}

void LCDclr (unsigned char d) {
HardwareSerial *hsp;
    hsp = LCDportarray[d];
    (*hsp).print("?f");     //clear command.
    if(d == 1) {
	lcd.clear();
    }
    delay(35); // < 34 and serial<->LCD converter freezes
}

void LCDquadpotMin (unsigned char d, unsigned char min) {
HardwareSerial *hsp;
    hsp = LCDportarray[d];
    (*hsp).print("?L");
    (*hsp).print(min);
}

void LCDquadpotMax (unsigned char d, unsigned char max) {
HardwareSerial *hsp;
    hsp = LCDportarray[d];
    (*hsp).print("?H");
    (*hsp).print(max);
}

void LCDquadpotPos (unsigned char d, unsigned char pos) {
HardwareSerial *hsp;
    hsp = LCDportarray[d];
    (*hsp).print("?Q");
    (*hsp).print(pos);
}

void LCDprintIP (unsigned char d, byte * ipa) {
HardwareSerial *hsp;
    hsp = LCDportarray[d];
    for(byte i = 0 ; i < 4 ; i++) {
	(*hsp).print(ipa[i]);
	if(i < 3) (*hsp).print('.');
	if(d == 1) {
	    lcd.print((int)ipa[i]);
	    if(i < 3) lcd.print('.');
	}
    }
}
