#include "alarm.h"
#define NTP_MAIN 1
#include "ntpstuff.h"
#if INTERNAL_WDT
// #include <avr/wdt.h>
#else
#include "wdt.h"
#endif
EthernetUDP Udp;

void (* resetFunc2) (void);
/*-------- NTP code ----------*/

const char pooln[] PROGMEM = "pool.ntp.org";
const char pool0[] PROGMEM = "0.ubuntu.pool.ntp.org";
const char pool1[] PROGMEM = "1.ubuntu.pool.ntp.org";
const char pool2[] PROGMEM = "2.ubuntu.pool.ntp.org";
const char pool3[] PROGMEM = "3.ubuntu.pool.ntp.org";
const char timba[] PROGMEM = "time-a.timefreq.bldrdoc.gov";
const char timbb[] PROGMEM = "time-b.timefreq.bldrdoc.gov";
const char timbc[] PROGMEM = "time-c.timefreq.bldrdoc.gov";
const char timna[] PROGMEM = "time-a.nist.gov";
const char timnb[] PROGMEM = "time-b.nist.gov";
const char timnc[] PROGMEM = "time-c.nist.gov";
const char timnd[] PROGMEM = "time-d.nist.gov";

const char * const ntp_hosts[] PROGMEM = {
    pooln, pool0, pool1, pool2, pool3,
    timba, timbb, timbc, timna, timnb, timnc, timnd
};

time_t getNTPtime(void) {
time_t result;
int j, bad_date_retries;
char ntpserver[30];
byte packetBuffer[NTP_PACKET_SIZE]; //buffer to hold incoming & outgoing packets

    Udp.begin(localUDPport);
    bad_date_retries = 3;
Retry:
    while (Udp.parsePacket() > 0) ; // discard any previously received packets
    for(j = 0 ; j < sizeof(ntp_hosts)/sizeof(char *) ; j++) {
#if NTPDEBUG
	DLFn("before sendNTPpacket()");
#endif
	strcpy_P(ntpserver, (char *)pgm_read_word(&(ntp_hosts[j])));
	DVFn("NTP Req", ntpserver);
	sendNTPpacket(ntpserver, packetBuffer);
#if NTPDEBUG
	DLFn("after sendNTPpacket()");
#endif
	uint32_t beginWait = millis();
	// DVn("beginWait ", beginWait);
	while (millis() - beginWait < 1500) {
	    // DVn("millis() ", millis());
	    int size = Udp.parsePacket();
	    if (size >= NTP_PACKET_SIZE) {
#if NTPDEBUG
		Serial.println(F("Receive NTP Response"));
#endif
		Udp.read(packetBuffer, NTP_PACKET_SIZE);  // read packet into the buffer
		unsigned long secsSince1900;
		// convert four bytes starting at location 40 to a long integer
		secsSince1900 =  (unsigned long)packetBuffer[40] << 24;
		secsSince1900 |= (unsigned long)packetBuffer[41] << 16;
		secsSince1900 |= (unsigned long)packetBuffer[42] << 8;
		secsSince1900 |= (unsigned long)packetBuffer[43];
		result = secsSince1900 - 2208988800UL;
		DVFn("time_t =", result);
		DVFn("year is", (int)year(result));
		if(year(result) > 2035 || year(result) < 2016) {
		    DVFn("bad date detected:", (int)year(result));
		    bad_date_retries++;
		    goto Retry;
		}
		ntp_fail_counter = 0;
		return (result);
	    }
#if INTERNAL_WDT
	    // wdt_reset();
#else
	    WDT_reset();
#endif
	}
#if NTPDEBUG
	DVFn("NTP timeout, j =", j);
#endif
    }
#if NTPDEBUG
    Serial.println(F("No Ethernet NTP Response :-("));
    delay(20);
#endif
    return 0; // return 0 if unable to get the time
}

void prepare_packet(byte *pbuffer) {
    // set all bytes in the buffer to 0
    memset(pbuffer, 0, NTP_PACKET_SIZE);
    // Initialize values needed to form NTP request
    // (see URL above for details on the packets)
    pbuffer[0] = 0b11100011;   // LI, Version, Mode
    pbuffer[1] = 0;     // Stratum, or type of clock
    pbuffer[2] = 6;     // Polling Interval
    pbuffer[3] = 0xEC;  // Peer Clock Precision
    // 8 bytes of zero for Root Delay & Root Dispersion
    pbuffer[12]  = 49;
    pbuffer[13]  = 0x4E;
    pbuffer[14]  = 49;
    pbuffer[15]  = 52;
}

// send an NTP request to the time server at the given host
void sendNTPpacket(char *host, byte *pbuffer) {
    prepare_packet(pbuffer);
    // all NTP fields have been given values, now
    // you can send a packet requesting a timestamp:
#if NTPDEBUG == 2
    DLFn("before Udp.beginPacket()");
#endif
    Udp.beginPacket(host, 123); //NTP requests are to port 123
#if NTPDEBUG == 2
    DLFn("before Udp.write()");
#endif
    Udp.write(pbuffer, NTP_PACKET_SIZE);
#if NTPDEBUG == 2
    DLFn("before Udp.endPacket()");
#endif
    Udp.endPacket();
}

#if NEEDED_NO_DNS_VERSION
// send an NTP request to the time server at the given address
void sendNTPpacketAddr(IPAddress &host, byte *pbuffer) {
    prepare_packet(pbuffer);
    // all NTP fields have been given values, now
    // you can send a packet requesting a timestamp:
#if NTPDEBUG
    DLFn("before Udp.beginPacket()");
#endif
    Udp.beginPacket(host, 123); //NTP requests are to port 123
#if NTPDEBUG
    DLFn("before Udp.write()");
#endif
    Udp.write(pbuffer, NTP_PACKET_SIZE);
#if NTPDEBUG
    DLFn("before Udp.endPacket()");
#endif
    Udp.endPacket();
}
#endif
